# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import os
import socket
import subprocess
import time
import uuid
from threading import Thread
from typing import AsyncGenerator, List, Mapping, Optional

import psutil
import requests
from jinja2.exceptions import TemplateError
from loguru import logger
from omegaconf import DictConfig, OmegaConf
from openai import APITimeoutError, AsyncStream, BadRequestError
from openai.types.chat import ChatCompletionChunk, ChatCompletionMessageParam
from pipecat.adapters.services.open_ai_adapter import OpenAILLMInvocationParams
from pipecat.frames.frames import (
    CancelFrame,
    EndFrame,
    LLMFullResponseEndFrame,
    LLMFullResponseStartFrame,
    LLMTextFrame,
)
from pipecat.processors.aggregators.openai_llm_context import OpenAILLMContext
from pipecat.services.openai.llm import OpenAILLMService
from transformers import AsyncTextIteratorStreamer, AutoModelForCausalLM, AutoTokenizer
from vllm.config import ModelConfig as vllmModelConfig

DEFAULT_GENERATION_KWARGS = {
    "max_new_tokens": 256,
    "temperature": 0.7,
    "top_p": 0.9,
    "do_sample": True,
}


class LLMUtilsMixin:
    """Utils for local LLM services."""

    def _maybe_add_user_message(self, messages: List[ChatCompletionMessageParam]) -> List[ChatCompletionMessageParam]:
        """
        Some LLMs like "nvidia/Llama-3.1-Nemotron-Nano-8B-v1" requires a user turn after the system prompt,
        this function is used to add a dummy user turn if the system prompt is followed by an assistant turn.
        """
        if len(messages) > 1 and messages[0]["role"] == "system" and messages[1]["role"] == "assistant":
            message = {"role": "user", "content": "Hi"}
            messages.insert(1, message)
        elif len(messages) == 1 and messages[0]["role"] == "system":
            messages.append({"role": "user", "content": "Hi"})
        return messages

    def _maybe_merge_consecutive_user_turns(
        self, messages: List[ChatCompletionMessageParam]
    ) -> List[ChatCompletionMessageParam]:
        """
        Merge consecutive user turns into a single turn,
        since some LLMs like "nvidia/Llama-3.1-Nemotron-Nano-8B-v1" do not support consecutive user turns.
        """
        if not messages:
            return messages

        merged_messages = []

        user_content = ""
        for message in messages:
            role = message["role"]
            if role != "user":
                # check if there's any preceeding user content, add them first
                if user_content:
                    merged_messages.append({"role": "user", "content": user_content})
                    user_content = ""
                merged_messages.append(message)
            else:
                if user_content:
                    user_content += "; " + message["content"]
                else:
                    user_content = message["content"]

        # add the last user content
        if user_content:
            merged_messages.append({"role": "user", "content": user_content})

        return merged_messages


class HuggingFaceLLMLocalService(LLMUtilsMixin):
    """
    HuggingFace LLM local service.
    """

    def __init__(
        self,
        model: str = "meta-llama/Meta-Llama-3-8B-Instruct",
        device: str = "cuda:0",
        dtype: str = "bfloat16",
        thinking_budget: int = 0,
        generation_kwargs: dict = None,
        apply_chat_template_kwargs: dict = None,
    ):
        self.device = device
        self.dtype = dtype
        self.thinking_budget = thinking_budget
        self.tokenizer = AutoTokenizer.from_pretrained(model)
        self.model = AutoModelForCausalLM.from_pretrained(
            model, device_map=device, dtype=dtype, trust_remote_code=True
        )  # type: AutoModelForCausalLM

        self.generation_kwargs = generation_kwargs if generation_kwargs else DEFAULT_GENERATION_KWARGS
        logger.debug(f"LLM generation kwargs: {self.generation_kwargs}")

        self.apply_chat_template_kwargs = apply_chat_template_kwargs if apply_chat_template_kwargs else {}
        if "tokenize" in self.apply_chat_template_kwargs:
            if self.apply_chat_template_kwargs["tokenize"] is not False:
                logger.warning(
                    f"Found `tokenize=True` in apply_chat_template_kwargs={self.apply_chat_template_kwargs},"
                    "it will be ignored and forced to `False`"
                )
            self.apply_chat_template_kwargs.pop("tokenize")

        logger.debug(f"LLM apply_chat_template kwargs: {self.apply_chat_template_kwargs}")

    def _apply_chat_template(self, messages: List[ChatCompletionMessageParam]) -> str:
        """
        Apply the chat template to the messages.
        """
        return self.tokenizer.apply_chat_template(messages, tokenize=False, **self.apply_chat_template_kwargs)

    def _get_prompt_from_messages(self, messages: List[ChatCompletionMessageParam]) -> str:
        """
        Get the formatted prompt from the conversation history messages.
        This function also tries to fix the messages if the LLM cannot handle consecutive turns of the same role,
        or requires a user turn after the system prompt.
        """
        try:
            prompt = self._apply_chat_template(messages)
            return prompt
        except TemplateError as e:
            logger.warning(f"Got TemplateError: {e}.")

        logger.debug(f"Input LLM messages: {messages}")
        if len(messages) > 1 and messages[0]["role"] == "system" and messages[1]["role"] == "assistant":
            logger.warning("Trying to fix by adding dummy user message after system prompt...")
            try:
                messages = self._maybe_add_user_message(messages)
                logger.debug(f"LLM messages after adding dummy user message: {messages}")
                prompt = self._apply_chat_template(messages)
                return prompt
            except TemplateError as e:
                logger.warning(f"Got TemplateError: {e}. Trying to fix by merging consecutive turns if possible.")

        try:
            new_messages = self._maybe_merge_consecutive_user_turns(messages)
            logger.debug(f"LLM messages after merging consecutive user turns: {new_messages}")
            prompt = self._apply_chat_template(new_messages)
            # Update the messages in place if successful
            messages.clear()
            messages.extend(new_messages)
            return prompt
        except Exception as e:
            logger.warning(f"Got Exception: {e}, messages: {messages}")
            raise e

    async def generate_stream(
        self, messages: List[ChatCompletionMessageParam], **kwargs
    ) -> AsyncGenerator[ChatCompletionChunk, None]:
        """
        Generate a stream of chat completion chunks from the messages.
        """
        # Convert messages to prompt format
        prompt = self._get_prompt_from_messages(messages)

        logger.debug(f"LLM prompt: {prompt}")

        inputs = self.tokenizer(prompt, add_special_tokens=False, return_tensors="pt").to(self.device)

        # Generate with streaming
        streamer = AsyncTextIteratorStreamer(self.tokenizer, skip_prompt=True, skip_special_tokens=True)
        generation_kwargs = {
            **inputs,
            "streamer": streamer,
            **self.generation_kwargs,
        }

        # Start generation in background
        thread = Thread(
            target=self.model.generate,
            kwargs=generation_kwargs,
        )
        thread.start()

        # Stream the output
        async for text in streamer:
            # logger.debug(f"Streamer yielded text: {text}")
            chunk = ChatCompletionChunk(
                id="hf-" + str(uuid.uuid4()),
                choices=[{"delta": {"content": text}, "finish_reason": None, "index": 0}],
                created=int(time.time()),
                model=self.model.config._name_or_path,
                object="chat.completion.chunk",
            )
            yield chunk


class HuggingFaceLLMService(OpenAILLMService):
    """
    LLM service that hosts a HuggingFace model.
    """

    def __init__(
        self,
        *,
        model: str = "google/gemma-7b-it",
        device: str = "cuda",
        dtype: str = "bfloat16",
        thinking_budget: int = 0,
        generation_kwargs: dict = None,
        apply_chat_template_kwargs: dict = None,
        **kwargs,
    ):
        self._model_name = model
        self._device = device
        self._dtype = dtype
        self._thinking_budget = thinking_budget
        self._generation_kwargs = generation_kwargs if generation_kwargs is not None else DEFAULT_GENERATION_KWARGS
        self._apply_chat_template_kwargs = apply_chat_template_kwargs if apply_chat_template_kwargs is not None else {}
        super().__init__(model=model, **kwargs)

    def create_client(self, api_key=None, base_url=None, **kwargs):
        """
        Create a HuggingFaceLLMLocalService client.
        """
        return HuggingFaceLLMLocalService(
            model=self._model_name,
            device=self._device,
            dtype=self._dtype,
            thinking_budget=self._thinking_budget,
            generation_kwargs=self._generation_kwargs,
            apply_chat_template_kwargs=self._apply_chat_template_kwargs,
        )

    async def _process_context(self, context: OpenAILLMContext):
        """Process a context through the LLM and push text frames.

        Args:
            context (OpenAILLMContext): The context to process, containing messages
                and other information needed for the LLM interaction.
        """
        await self.push_frame(LLMFullResponseStartFrame())
        cumulative_text = ""
        try:
            await self.start_ttfb_metrics()
            messages = context.get_messages()
            async for chunk in self._client.generate_stream(messages):
                if chunk.choices[0].delta.content:
                    await self.stop_ttfb_metrics()
                    text = chunk.choices[0].delta.content
                    cumulative_text += text
                    frame = LLMTextFrame(text)
                    await self.push_frame(frame)
        except Exception as e:
            logger.error(f"Error in _process_context: {e}", exc_info=True)
            raise
        finally:
            cumulative_text = " ".join(cumulative_text.split()).strip()
            if not cumulative_text:
                logger.warning(f"LLM response is empty for context: {context}")
            await self.push_frame(LLMFullResponseEndFrame())

    async def get_chat_completions(
        self, params_from_context: OpenAILLMInvocationParams
    ) -> AsyncGenerator[ChatCompletionChunk, None]:
        """Create a streaming chat completion using HuggingFace model.

        Args:
            context (OpenAILLMContext): The context object containing tools configuration
                and other settings for the chat completion.
            messages (List[ChatCompletionMessageParam]): The list of messages comprising
                the conversation history and current request.

        Returns:
            AsyncGenerator[ChatCompletionChunk]: A streaming response of chat completion
                chunks that can be processed asynchronously.
        """
        messages = params_from_context["messages"]

        return self._client.generate_stream(messages)


class VLLMService(OpenAILLMService, LLMUtilsMixin):
    """
    LLM service that hosts a vLLM server.
    """

    def __init__(
        self,
        *,
        model: str,
        device: str = "cuda",
        api_key="None",
        base_url="http://localhost:8000/v1",
        organization="None",
        project="None",
        default_headers: Optional[Mapping[str, str]] = None,
        params: Optional[OpenAILLMService.InputParams] = None,
        thinking_budget: int = 0,
        start_vllm_on_init: bool = False,
        vllm_server_params: Optional[str] = None,
        vllm_server_max_wait_time: int = 3600,  # 1 hour max wait time
        vllm_server_check_interval: int = 5,  # check server every 5 seconds
        **kwargs,
    ):
        self._device = device
        self._vllm_server_max_wait_time = vllm_server_max_wait_time
        self._vllm_server_check_interval = vllm_server_check_interval
        if start_vllm_on_init:
            base_url = self._start_vllm_server(model, vllm_server_params, base_url)

        super().__init__(
            model=model,
            api_key=api_key,
            base_url=base_url,
            organization=organization,
            project=project,
            default_headers=default_headers,
            params=params,
            **kwargs,
        )
        self._thinking_budget = thinking_budget
        self._vllm_server_params = vllm_server_params
        self._start_vllm_on_init = start_vllm_on_init

        # TODO: handle thinking budget
        logger.info(
            f"VLLMService initialized with model: {model}, api_key: {api_key}, base_url: {base_url},"
            f"params: {params}, thinking_budget: {thinking_budget}"
        )

    def _start_vllm_server(
        self, model: str, vllm_server_params: Optional[str] = None, base_url: Optional[str] = None
    ) -> str:
        """
        Start a vllm server and return the base url.
        """

        requested_port = None
        # If base_url is provided, extract port from it
        if base_url:
            try:
                # Extract port from base_url like "http://localhost:8003/v1"
                from urllib.parse import urlparse

                parsed_url = urlparse(base_url)
                if parsed_url.port:
                    requested_port = parsed_url.port
            except Exception as e:
                logger.warning(
                    f"Could not parse port from base_url {base_url}: {e}, using port from vllm_server_params"
                )

        # Parse port from vllm_server_params, default to 8000
        if vllm_server_params:
            params_list = vllm_server_params.split()
            for i, param in enumerate(params_list):
                if param == "--port" and i + 1 < len(params_list):
                    try:
                        param_port = int(params_list[i + 1])
                        if requested_port is None:
                            requested_port = param_port
                        else:
                            if param_port != requested_port:
                                logger.warning(
                                    f"Port {param_port} from vllm_server_params is different from base_url port"
                                    f"{requested_port}, using new port {param_port}"
                                )
                                requested_port = param_port
                        break
                    except ValueError:
                        logger.warning(f"Invalid port number: {params_list[i + 1]}, using default 8000")

        if requested_port is None:
            # try to use default port
            requested_port = 8000

        def find_available_port(start_port: int) -> int:
            """Find an available port starting from start_port"""
            for port in range(start_port, start_port + 100):  # Try up to 100 ports
                try:
                    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                        s.bind(('localhost', port))
                        return port
                except OSError:
                    continue
            raise RuntimeError(f"Could not find an available port starting from {start_port}")

        def get_pid_on_port(port: int) -> Optional[int]:
            for conn in psutil.net_connections(kind="inet"):
                if conn.laddr.port == port and conn.status == psutil.CONN_LISTEN:
                    return conn.pid
            return None

        def check_server_model(port: int, verbose: bool = False) -> tuple[bool, str]:
            """Check if server is running on port and return (is_running, model_name)"""
            try:
                response = requests.get(f"http://localhost:{port}/v1/models", timeout=5)
                if response.status_code == 200:
                    # get the PID for the server process
                    pid = get_pid_on_port(port)
                    if pid is not None and verbose:
                        logger.warning(
                            f"Found vLLM server process (PID: {pid}) on port {port}, you can use `lsof -i :{port}`"
                            "to find the process and kill it if you want to start a new server."
                        )
                    models_data = response.json()
                    if "data" in models_data and models_data["data"]:
                        served_model = models_data["data"][0].get("id", "")
                        return True, served_model
                    return True, ""
                return False, ""
            except (requests.exceptions.RequestException, requests.exceptions.Timeout):
                return False, ""

        # First, check if vLLM server is already running on the requested port
        is_running, served_model = check_server_model(requested_port, verbose=True)
        if is_running:
            if served_model == model:
                final_base_url = f"http://localhost:{requested_port}/v1"
                logger.info(f"vLLM server is already running at {final_base_url} with the correct model: {model}")
                return final_base_url
            else:
                logger.warning(
                    f"vLLM server on port {requested_port} is serving model '{served_model}' but we need '{model}'."
                    "Finding new port..."
                )

        # Find an available port for our model
        port = find_available_port(requested_port)
        if port != requested_port:
            logger.info(f"Using port {port} instead of requested port {requested_port}")

        final_base_url = f"http://localhost:{port}/v1"

        # Check if there's already a vLLM process running on the same port and model
        for proc in psutil.process_iter(['pid', 'name', 'cmdline']):
            try:
                if proc.info['cmdline'] and any('vllm' in arg and 'serve' in arg for arg in proc.info['cmdline']):
                    # Check if this process is using the same port and model
                    cmdline_str = ' '.join(proc.info['cmdline'])
                    if f"--port {port}" in cmdline_str:
                        # Extract the model from the command line
                        cmdline_parts = proc.info['cmdline']
                        model_index = -1
                        for i, arg in enumerate(cmdline_parts):
                            if arg == "serve" and i + 1 < len(cmdline_parts):
                                model_index = i + 1
                                break

                        if model_index != -1 and model_index < len(cmdline_parts):
                            running_model = cmdline_parts[model_index]
                            if running_model == model:
                                logger.info(
                                    f"Found existing vLLM server process (PID: {proc.info['pid']}) on port {port}"
                                    f"serving model {model}"
                                )
                                # Wait a bit and check if it's responding
                                time.sleep(2)
                                is_running, served_model = check_server_model(port)
                                if is_running and served_model == model:
                                    logger.info(
                                        f"Existing vLLM server is responding at {final_base_url} with correct model"
                                    )
                                    return final_base_url
                                else:
                                    logger.warning(
                                        f"Existing vLLM process found on port {port} but not responding correctly,"
                                        "will start new server"
                                    )
                            else:
                                logger.info(
                                    f"Found vLLM process on port {port} but serving different model '{running_model}'"
                                    f"(need '{model}'). Will start new server."
                                )
            except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
                continue

        # Build the command with the determined port
        cmd_parts = ["vllm", "serve", model]

        # Parse and modify vllm_server_params to use the correct port
        if vllm_server_params:
            # parse the vllm_server_params and add the port to the command
            params_list = vllm_server_params.split()
            modified_params = []
            i = 0
            while i < len(params_list):
                if params_list[i] == "--port" and i + 1 < len(params_list):
                    # Replace the port with our determined port
                    modified_params.extend(["--port", str(port)])
                    i += 2  # Skip the original port value
                else:
                    modified_params.append(params_list[i])
                    i += 1
            cmd_parts.extend(modified_params)
        else:
            # Add port if vllm_server_params is not provided
            cmd_parts.extend(["--port", str(port)])

        logger.info(f"Starting vLLM server with command: {' '.join(cmd_parts)}")
        logger.warning("It will take a while to download the model if it's not already downloaded.")
        # Set up environment variables for device configuration
        env = os.environ.copy()
        if self._device and self._device != "cpu":
            # Extract CUDA device number if it's in format "cuda:0", "cuda:1", etc.
            if self._device.startswith("cuda:"):
                device_id = self._device.split(":")[1]
                env["CUDA_VISIBLE_DEVICES"] = device_id
                logger.info(f"Setting CUDA_VISIBLE_DEVICES={device_id}")
            elif self._device == "cuda":
                # Use default CUDA device (don't set CUDA_VISIBLE_DEVICES)
                logger.info("Using default CUDA device")
            else:
                # For other device strings, try to extract device number
                logger.warning(f"Unknown device format: {self._device}, using as-is")
                env["CUDA_VISIBLE_DEVICES"] = self._device
        elif self._device == "cpu":
            env["CUDA_VISIBLE_DEVICES"] = ""
            logger.info("Setting CUDA_VISIBLE_DEVICES='' to use CPU")

        try:
            # Start the vLLM server process with environment variables
            process = subprocess.Popen(
                cmd_parts,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                env=env,
                preexec_fn=os.setsid if os.name != 'nt' else None,  # Create new process group
            )

            # Store the process for potential cleanup later
            self._vllm_process = process

            # Wait for server to start up
            max_wait_time = self._vllm_server_max_wait_time
            check_interval = self._vllm_server_check_interval
            waited_time = 0

            logger.info(f"Waiting for vLLM server to start on port {port}...")
            while waited_time < max_wait_time:
                is_running, served_model = check_server_model(port)
                if is_running and served_model == model:
                    logger.info(f"vLLM server started successfully at {final_base_url} serving model: {model}")
                    return final_base_url
                elif is_running and served_model != model:
                    logger.warning(
                        f"vLLM server started but serving wrong model '{served_model}' instead of '{model}'."
                        "Continuing to wait..."
                    )

                # Check if process is still running
                if process.poll() is not None:
                    # Process has terminated
                    stdout, stderr = process.communicate()
                    logger.error(f"vLLM server process terminated unexpectedly. stdout: {stdout}, stderr: {stderr}")
                    raise RuntimeError(f"Failed to start vLLM server: {stderr}")

                time.sleep(check_interval)
                waited_time += check_interval
                logger.debug(f"Still waiting for vLLM server on port {port}... ({waited_time}s)")

            # If we get here, server didn't start in time
            logger.error(f"vLLM server failed to start within {max_wait_time} seconds on port {port}")
            process.terminate()
            raise RuntimeError(f"vLLM server failed to start within {max_wait_time} seconds on port {port}")

        except FileNotFoundError:
            logger.error("vLLM not found. Please install vLLM: pip install vllm")
            raise RuntimeError("vLLM not found. Please install vLLM: pip install vllm")
        except Exception as e:
            logger.error(f"Failed to start vLLM server: {e}")
            self._stop_vllm_server()
            raise e

    def _stop_vllm_server(self):
        """Stop the vLLM server process if it's running."""
        if hasattr(self, '_vllm_process') and self._vllm_process:
            logger.info(f"Stopping vLLM server process {self._vllm_process.pid}")
            self._vllm_process.terminate()

    async def stop(self, frame: EndFrame):
        """Stop the LLM service.

        Args:
            frame: The end frame.
        """
        await super().stop(frame)
        self._stop_vllm_server()

    async def cancel(self, frame: CancelFrame):
        """Cancel the LLM service.

        Args:
            frame: The cancel frame.
        """
        await super().cancel(frame)
        self._stop_vllm_server()

    async def get_chat_completions(
        self, params_from_context: OpenAILLMInvocationParams
    ) -> AsyncStream[ChatCompletionChunk]:
        """Get streaming chat completions from OpenAI API.

        Args:
            context: The LLM context containing tools and configuration.
            messages: List of chat completion messages to send.

        Returns:
            Async stream of chat completion chunks.
        """

        params = self.build_chat_completion_params(params_from_context)
        messages = params_from_context["messages"]
        if self._retry_on_timeout:
            try:
                chunks = await asyncio.wait_for(
                    self._get_response_from_client(messages, params), timeout=self._retry_timeout_secs
                )
                return chunks
            except (APITimeoutError, asyncio.TimeoutError):
                # Retry, this time without a timeout so we get a response
                logger.debug(f"{self}: Retrying chat completion due to timeout")
                chunks = await self._get_response_from_client(messages, params)
                return chunks
        else:
            chunks = await self._get_response_from_client(messages, params)
            return chunks

    async def _get_response_from_client(
        self, messages: List[ChatCompletionMessageParam], params: dict
    ) -> AsyncStream[ChatCompletionChunk]:
        """Get a response from the client."""

        try:
            chunks = await self._client.chat.completions.create(**params)
        except BadRequestError as e:
            logger.error(f"Error in _get_response_from_client: {e}, trying to fix...")
            logger.debug(f"LLM messages before fixing: {messages}")
            messages = self._maybe_add_user_message(messages)
            messages = self._maybe_merge_consecutive_user_turns(messages)
            logger.debug(f"LLM messages after fixing: {messages}")
            params["messages"] = messages
            chunks = await self._client.chat.completions.create(**params)

        return chunks


def get_llm_service_from_config(config: DictConfig) -> OpenAILLMService:
    """Get an LLM service from the configuration."""
    backend = config.type

    # If backend is "auto", try to detect the best backend
    if backend == "auto":
        model_name = config.get("model")
        if not model_name:
            raise ValueError("Model name is required for LLM")

        try:
            _ = vllmModelConfig(model_name, trust_remote_code=True)
            backend = "vllm"
            logger.info(f"Auto-detected vLLM as the best backend for model {model_name}")
        except Exception as e:
            logger.info(
                f"The LLM doesn't seem to be supported by vLLM yet (error: {e}), using HuggingFace as the backend"
                f"for model: {model_name}. If you are sure that the LLM is supported by vLLM, you can set `type: vllm`"
                "in the config file to force using vLLM."
            )
            backend = "hf"

    assert backend in [
        "hf",
        "vllm",
        "auto",
    ], f"Invalid backend: {backend}, only `hf`, `vllm`, and `auto` are supported."

    if backend == "hf":
        llm_model = config.model
        llm_device = config.device
        llm_dtype = config.dtype
        llm_generation_kwargs = config.get("generation_kwargs", {})
        if llm_generation_kwargs is not None:
            llm_generation_kwargs = OmegaConf.to_container(llm_generation_kwargs, resolve=True)
        llm_apply_chat_template_kwargs = config.get("apply_chat_template_kwargs", None)
        if llm_apply_chat_template_kwargs is not None:
            llm_apply_chat_template_kwargs = OmegaConf.to_container(llm_apply_chat_template_kwargs, resolve=True)
        llm_thinking_budget = config.get("thinking_budget", 0)
        return HuggingFaceLLMService(
            model=llm_model,
            device=llm_device,
            dtype=llm_dtype,
            generation_kwargs=llm_generation_kwargs,
            apply_chat_template_kwargs=llm_apply_chat_template_kwargs,
            thinking_budget=llm_thinking_budget,
        )
    elif backend == "vllm":
        llm_model = config.get("model", "vllm_server")
        llm_api_key = config.get("api_key", "None")
        llm_base_url = config.get("base_url", "http://localhost:8000/v1")
        llm_organization = config.get("organization", "None")
        llm_project = config.get("project", "None")
        llm_default_headers = config.get("default_headers", None)
        llm_params = config.get("vllm_generation_params", None)
        llm_dtype = config.dtype
        vllm_server_params = config.get("vllm_server_params", None)
        if vllm_server_params is not None:
            if "dtype" not in vllm_server_params:
                vllm_server_params = f"--dtype {llm_dtype} {vllm_server_params}"
                logger.info(f"Adding dtype {llm_dtype} to vllm_server_params: {vllm_server_params}")
        if llm_params is not None:
            # cast into OpenAILLMService.InputParams object
            llm_params = OmegaConf.to_container(llm_params, resolve=True)
            extra = llm_params.get("extra", None)
            # ensure extra is a dictionary
            if extra is None:
                llm_params["extra"] = {}
            elif not isinstance(extra, dict):
                raise ValueError(f"extra must be a dictionary, got {type(extra)}")
            llm_params = OpenAILLMService.InputParams(**llm_params)
        else:
            llm_params = OpenAILLMService.InputParams()
        llm_thinking_budget = config.get("thinking_budget", 0)
        return VLLMService(
            model=llm_model,
            api_key=llm_api_key,
            base_url=llm_base_url,
            organization=llm_organization,
            project=llm_project,
            default_headers=llm_default_headers,
            params=llm_params,
            thinking_budget=llm_thinking_budget,
            start_vllm_on_init=config.get("start_vllm_on_init", False),
            vllm_server_params=vllm_server_params,
        )
    else:
        raise ValueError(f"Invalid LLM backend: {backend}")



# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
from collections import Counter
from collections import OrderedDict as od
from pathlib import Path
from typing import Dict, List, Union

import librosa
import numpy as np

from nemo.collections.asr.parts.utils.speaker_utils import (
    audio_rttm_map,
    get_subsegments_scriptable,
    get_uniqname_from_filepath,
    rttm_to_labels,
    segments_manifest_to_subsegments_manifest,
    write_rttm2manifest,
)
from nemo.utils import logging
from nemo.utils.data_utils import DataStoreObject


def get_rounded_str_float(num: float, output_precision: int, min_precision=1, max_precision=3) -> str:
    """
    Get a string of a float number with rounded precision.

    Args:
        num (float): float number to round
        output_precision (int): precision of the output floating point number
        min_precision (int, optional): Minimum precision of the output floating point number. Defaults to 1.
        max_precision (int, optional): Maximum precision of the output floating point number. Defaults to 3.

    Returns:
        (str): Return a string of a float number with rounded precision.
    """
    output_precision = min(max_precision, max(min_precision, output_precision))
    return f"{num:.{output_precision}f}"


def get_ctm_line(
    source: str,
    channel: int,
    start_time: float,
    duration: float,
    token: str,
    conf: float,
    type_of_token: str,
    speaker: str,
    NA_token: str = 'NA',
    UNK: str = 'unknown',
    default_channel: str = '1',
    output_precision: int = 2,
) -> str:
    """
    Get a line in Conversation Time Mark (CTM) format. Following CTM format appeared in
    `Rich Transcription Meeting Eval Plan: RT09` document.

    CTM Format:
        <SOURCE><SP><CHANNEL><SP><BEG-TIME><SP><DURATION><SP><TOKEN><SP><CONF><SP><TYPE><SP><SPEAKER><NEWLINE>

    Reference:
        https://web.archive.org/web/20170119114252/
        http://www.itl.nist.gov/iad/mig/tests/rt/2009/docs/rt09-meeting-eval-plan-v2.pdf

    Args:
        source (str): <SOURCE> is name of the source file, session name or utterance ID
        channel (int): <CHANNEL> is channel number defaults to 1
        start_time (float): <BEG_TIME> is the begin time of the word, which we refer to as `start_time` in NeMo.
        duration (float): <DURATION> is duration of the word
        token (str): <TOKEN> Token or word for the current entry
        conf (float): <CONF> is a floating point number between 0 (no confidence) and 1 (certainty).
                      A value of “NA” is used (in CTM format data)
                      when no confidence is computed and in the reference data.
        type_of_token (str): <TYPE> is the token type. The legal values of <TYPE> are
                      “lex”, “frag”, “fp”, “un-lex”, “for-lex”, “non-lex”, “misc”, or “noscore”
        speaker (str): <SPEAKER> is a string identifier for the speaker who uttered the token.
                      This should be “null” for non-speech tokens and “unknown” when
                      the speaker has not been determined.
        NA_token (str, optional): A token for  . Defaults to '<NA>'.
        output_precision (int, optional): The precision of the output floating point number. Defaults to 3.

    Returns:
        str: Return a line in CTM format filled with the given information.
    """
    VALID_TOKEN_TYPES = ["lex", "frag", "fp", "un-lex", "for-lex", "non-lex", "misc", "noscore"]

    if type(start_time) == str and start_time.replace('.', '', 1).isdigit():
        start_time = float(start_time)
    elif type(start_time) != float:
        raise ValueError(f"`start_time` must be a float or str containing float, but got {type(start_time)}")

    if type(duration) == str and duration.replace('.', '', 1).isdigit():
        duration = float(duration)
    elif type(duration) != float:
        raise ValueError(f"`duration` must be a float or str containing float, but got {type(duration)}")

    if type(conf) == str and conf.replace('.', '', 1).isdigit():
        conf = float(conf)
    elif conf is None:
        conf = NA_token
    elif type(conf) != float:
        raise ValueError(f"`conf` must be a float or str containing float, but got {type(conf)}")

    if channel is not None and type(channel) != int:
        channel = str(channel)
    if conf is not None and type(conf) == float and not (0 <= conf <= 1):
        raise ValueError(f"`conf` must be between 0 and 1, but got {conf}")
    if type_of_token is not None and type(type_of_token) != str:
        raise ValueError(f"`type` must be a string, but got {type(type_of_token)} type {type_of_token}")
    if type_of_token is not None and type_of_token not in VALID_TOKEN_TYPES:
        raise ValueError(f"`type` must be one of {VALID_TOKEN_TYPES}, but got {type_of_token} type {type_of_token}")
    if speaker is not None and type(speaker) != str:
        raise ValueError(f"`speaker` must be a string, but got {type(speaker)}")

    channel = default_channel if channel is None else channel
    conf = NA_token if conf is None else conf
    speaker = NA_token if speaker is None else speaker
    type_of_token = UNK if type_of_token is None else type_of_token
    start_time = get_rounded_str_float(start_time, output_precision)
    duration = get_rounded_str_float(duration, output_precision)
    conf = get_rounded_str_float(conf, output_precision) if conf != NA_token else conf
    return f"{source} {channel} {start_time} {duration} {token} {conf} {type_of_token} {speaker}\n"


def rreplace(s: str, old: str, new: str) -> str:
    """
    Replace end of string.

    Args:
        s (str): string to operate on
        old (str): ending of string to replace
        new (str): replacement for ending of string
    Returns:
        new.join(li) (string): new string with end replaced
    """
    li = s.rsplit(old, 1)
    return new.join(li)


def get_uniq_id_with_period(path: str) -> str:
    """
    Get uniq_id from path string with period in it.

    Args:
        path (str): path to audio file
    Returns:
        uniq_id (str): unique speaker ID
    """
    split_path = os.path.basename(path).split('.')[:-1]
    uniq_id = '.'.join(split_path) if len(split_path) > 1 else split_path[0]
    return uniq_id


def get_subsegment_dict(subsegments_manifest_file: str, window: float, shift: float, deci: int) -> Dict[str, dict]:
    """
    Get subsegment dictionary from manifest file.

    Args:
        subsegments_manifest_file (str): Path to subsegment manifest file
        window (float): Window length for segmentation
        shift (float): Shift length for segmentation
        deci (int): Rounding number of decimal places
    Returns:
        _subsegment_dict (dict): Subsegment dictionary
    """
    _subsegment_dict = {}
    with open(subsegments_manifest_file, 'r') as subsegments_manifest:
        segments = subsegments_manifest.readlines()
        for segment in segments:
            segment = segment.strip()
            dic = json.loads(segment)
            audio, offset, duration = dic['audio_filepath'], dic['offset'], dic['duration']
            subsegments = get_subsegments_scriptable(offset=offset, window=window, shift=shift, duration=duration)
            if dic['uniq_id'] is not None:
                uniq_id = dic['uniq_id']
            else:
                uniq_id = get_uniq_id_with_period(audio)
            if uniq_id not in _subsegment_dict:
                _subsegment_dict[uniq_id] = {'ts': [], 'json_dic': []}
            for subsegment in subsegments:
                start, dur = subsegment
            _subsegment_dict[uniq_id]['ts'].append([round(start, deci), round(start + dur, deci)])
            _subsegment_dict[uniq_id]['json_dic'].append(dic)
    return _subsegment_dict


def get_input_manifest_dict(input_manifest_path: str) -> Dict[str, dict]:
    """
    Get dictionary from manifest file.

    Args:
        input_manifest_path (str): Path to manifest file
    Returns:
        input_manifest_dict (dict): Dictionary from manifest file
    """
    input_manifest_dict = {}
    with open(input_manifest_path, 'r') as input_manifest_fp:
        json_lines = input_manifest_fp.readlines()
        for json_line in json_lines:
            dic = json.loads(json_line)
            dic["text"] = "-"
            uniq_id = get_uniqname_from_filepath(dic["audio_filepath"])
            input_manifest_dict[uniq_id] = dic
    return input_manifest_dict


def write_truncated_subsegments(
    input_manifest_dict: Dict[str, dict],
    _subsegment_dict: Dict[str, dict],
    output_manifest_path: str,
    step_count: int,
    deci: int,
):
    """
    Write subsegments to manifest filepath.

    Args:
        input_manifest_dict (dict): Input manifest dictionary
        _subsegment_dict (dict): Input subsegment dictionary
        output_manifest_path (str): Path to output manifest file
        step_count (int): Number of the unit segments you want to create per utterance
        deci (int): Rounding number of decimal places
    """
    with open(output_manifest_path, 'w') as output_manifest_fp:
        for uniq_id, subseg_dict in _subsegment_dict.items():
            subseg_array = np.array(subseg_dict['ts'])
            subseg_array_idx = np.argsort(subseg_array, axis=0)
            chunked_set_count = subseg_array_idx.shape[0] // step_count

            for idx in range(chunked_set_count - 1):
                chunk_index_stt = subseg_array_idx[:, 0][idx * step_count]
                chunk_index_end = subseg_array_idx[:, 1][(idx + 1) * step_count]
                offset_sec = subseg_array[chunk_index_stt, 0]
                end_sec = subseg_array[chunk_index_end, 1]
                dur = round(end_sec - offset_sec, deci)
                meta = input_manifest_dict[uniq_id]
                meta['offset'] = offset_sec
                meta['duration'] = dur
                json.dump(meta, output_manifest_fp)
                output_manifest_fp.write("\n")


def write_file(name: str, lines: List[dict], idx: int):
    """
    Write json lines to file.

    Args:
        name (str): Output file path
        lines (list): List of json lines
        idx (int): Indices to dump to the file
    """
    with open(name, 'w') as fout:
        for i in idx:
            dic = lines[i]
            json.dump(dic, fout)
            fout.write('\n')


def read_file(pathlist: str) -> List[str]:
    """
    Read list of lines from target file.

    Args:
        pathlist (str): Input file path
    Returns:
        sorted(pathlist) (list): List of lines
    """
    with open(pathlist, 'r') as f:
        pathlist = f.readlines()
    return sorted(pathlist)


def get_dict_from_wavlist(pathlist: List[str]) -> Dict[str, str]:
    """
    Read dictionaries from list of lines

    Args:
        pathlist (list): List of file paths
    Returns:
        path_dict (dict): Dictionary containing dictionaries read from files
    """
    path_dict = od()
    pathlist = sorted(pathlist)
    for line_path in pathlist:
        uniq_id = os.path.basename(line_path).split('.')[0]
        path_dict[uniq_id] = line_path
    return path_dict


def get_dict_from_list(data_pathlist: List[str], uniqids: List[str]) -> Dict[str, str]:
    """
    Create dictionaries from list of lines

    Args:
        data_pathlist (list): List of file paths
        uniqids (list): List of file IDs
    Returns:
        path_dict (dict): Dictionary containing file paths
    """
    path_dict = {}
    for line_path in data_pathlist:
        uniq_id = os.path.basename(line_path).split('.')[0]
        if uniq_id in uniqids:
            path_dict[uniq_id] = line_path
        else:
            raise ValueError(f'uniq id {uniq_id} is not in wav filelist')
    return path_dict


def get_path_dict(data_path: str, uniqids: List[str], len_wavs: int = None) -> Dict[str, str]:
    """
    Create dictionary from list of lines (using the get_dict_from_list function)

    Args:
        data_path (str): Path to file containing list of files
        uniqids (list): List of file IDs
        len_wavs (int): Length of file list
    Returns:
        data_pathdict (dict): Dictionary containing file paths
    """
    if data_path is not None:
        data_pathlist = read_file(data_path)
        if len_wavs is not None:
            assert len(data_pathlist) == len_wavs
            data_pathdict = get_dict_from_list(data_pathlist, uniqids)
    elif len_wavs is not None:
        data_pathdict = {uniq_id: None for uniq_id in uniqids}
    return data_pathdict


def create_segment_manifest(
    input_manifest_path: str, output_manifest_path: str, window: float, shift: float, step_count: int, deci: int
):
    """
    Create segmented manifest file from base manifest file

    Args:
        input_manifest_path (str): Path to input manifest file
        output_manifest_path (str): Path to output manifest file
        window (float): Window length for segmentation
        shift (float): Shift length for segmentation
        step_count (int): Number of the unit segments you want to create per utterance
        deci (int): Rounding number of decimal places
    """
    if '.json' not in input_manifest_path:
        raise ValueError("input_manifest_path file should be .json file format")
    if output_manifest_path and '.json' not in output_manifest_path:
        raise ValueError("output_manifest_path file should be .json file format")
    elif not output_manifest_path:
        output_manifest_path = rreplace(input_manifest_path, '.json', f'_{step_count}seg.json')

    input_manifest_dict = get_input_manifest_dict(input_manifest_path)
    segment_manifest_path = rreplace(input_manifest_path, '.json', '_seg.json')
    subsegment_manifest_path = rreplace(input_manifest_path, '.json', '_subseg.json')
    min_subsegment_duration = 0.05
    step_count = int(step_count)

    AUDIO_RTTM_MAP = audio_rttm_map(input_manifest_path)
    segments_manifest_file = write_rttm2manifest(AUDIO_RTTM_MAP, segment_manifest_path, deci)
    subsegments_manifest_file = subsegment_manifest_path
    segments_manifest_to_subsegments_manifest(
        segments_manifest_file,
        subsegments_manifest_file,
        window,
        shift,
        min_subsegment_duration,
    )
    subsegments_dict = get_subsegment_dict(subsegments_manifest_file, window, shift, deci)
    write_truncated_subsegments(input_manifest_dict, subsegments_dict, output_manifest_path, step_count, deci)
    os.remove(segment_manifest_path)
    os.remove(subsegment_manifest_path)


def create_manifest(
    wav_path: str,
    manifest_filepath: str,
    text_path: str = None,
    rttm_path: str = None,
    uem_path: str = None,
    ctm_path: str = None,
    add_duration: bool = False,
):
    """
    Create base manifest file

    Args:
        wav_path (str): Path to list of wav files
        manifest_filepath (str): Path to output manifest file
        text_path (str): Path to list of text files
        rttm_path (str): Path to list of rttm files
        uem_path (str): Path to list of uem files
        ctm_path (str): Path to list of ctm files
        add_duration (bool): Whether to add durations to the manifest file
    """
    if os.path.exists(manifest_filepath):
        os.remove(manifest_filepath)
    wav_pathlist = read_file(wav_path)
    wav_pathdict = get_dict_from_wavlist(wav_pathlist)
    len_wavs = len(wav_pathlist)
    uniqids = sorted(wav_pathdict.keys())

    text_pathdict = get_path_dict(text_path, uniqids, len_wavs)
    rttm_pathdict = get_path_dict(rttm_path, uniqids, len_wavs)
    uem_pathdict = get_path_dict(uem_path, uniqids, len_wavs)
    ctm_pathdict = get_path_dict(ctm_path, uniqids, len_wavs)

    lines = []
    for uid in uniqids:
        wav, text, rttm, uem, ctm = (
            wav_pathdict[uid],
            text_pathdict[uid],
            rttm_pathdict[uid],
            uem_pathdict[uid],
            ctm_pathdict[uid],
        )

        audio_line = wav.strip()
        if rttm is not None:
            rttm = rttm.strip()
            labels = rttm_to_labels(rttm)
            num_speakers = Counter([label.split()[-1] for label in labels]).keys().__len__()
        else:
            num_speakers = None

        if uem is not None:
            uem = uem.strip()

        if text is not None:
            with open(text.strip()) as f:
                text = f.readlines()[0].strip()
        else:
            text = "-"

        if ctm is not None:
            ctm = ctm.strip()

        duration = None
        if add_duration:
            y, sr = librosa.load(audio_line, sr=None)
            duration = librosa.get_duration(y=y, sr=sr)
        meta = [
            {
                "audio_filepath": audio_line,
                "offset": 0,
                "duration": duration,
                "label": "infer",
                "text": text,
                "num_speakers": num_speakers,
                "rttm_filepath": rttm,
                "uem_filepath": uem,
                "ctm_filepath": ctm,
            }
        ]
        lines.extend(meta)

    write_file(manifest_filepath, lines, range(len(lines)))


def read_manifest(manifest: Union[Path, str]) -> List[dict]:
    """
    Read manifest file

    Args:
        manifest (str or Path): Path to manifest file
    Returns:
        data (list): List of JSON items
    """
    manifest = DataStoreObject(str(manifest))

    data = []
    try:
        f = open(manifest.get(), 'r', encoding='utf-8')
    except:
        raise Exception(f"Manifest file could not be opened: {manifest}")

    errors = []
    for line in f.readlines():
        line = line.strip()
        if not line:
            continue
        try:
            item = json.loads(line)
        except json.JSONDecodeError:
            errors.append(line)
            continue
        data.append(item)
    f.close()
    if errors:
        logging.error(f"{len(errors)} Errors encountered while reading manifest file: {manifest}")
        for error in errors:
            logging.error(f"-- Failed to parse line: `{error}`")
        raise RuntimeError(f"Errors encountered while reading manifest file: {manifest}")
    return data


def write_manifest(output_path: Union[Path, str], target_manifest: List[dict], ensure_ascii: bool = True):
    """
    Write to manifest file

    Args:
        output_path (str or Path): Path to output manifest file
        target_manifest (list): List of manifest file entries
        ensure_ascii (bool): default is True, meaning the output is guaranteed to have all incoming
                             non-ASCII characters escaped. If ensure_ascii is false, these characters
                             will be output as-is.
    """
    with open(output_path, "w", encoding="utf-8") as outfile:
        for tgt in target_manifest:
            json.dump(tgt, outfile, ensure_ascii=ensure_ascii)
            outfile.write('\n')


def write_ctm(output_path: str, target_ctm: Dict[str, dict]):
    """
    Write ctm entries from diarization session to a .ctm file.

    Args:
        output_path (str): target file path
        target_ctm (dict): list of ctm entries
    """
    target_ctm.sort(key=lambda y: y[0])
    with open(output_path, "w") as outfile:
        for pair in target_ctm:
            tgt = pair[1]
            outfile.write(tgt)


def write_text(output_path: str, target_ctm: Dict[str, dict]):
    """
    Write text from diarization session to a .txt file

    Args:
        output_path (str): target file path
        target_ctm (dict): list of ctm entries
    """
    target_ctm.sort(key=lambda y: y[0])
    with open(output_path, "w") as outfile:
        for pair in target_ctm:
            tgt = pair[1]
            word = tgt.split(' ')[4]
            outfile.write(word + ' ')
        outfile.write('\n')


def filepath_to_absolute(filepath: str | Path, base_path: Path) -> Path:
    """
    Return absolute path to an audio file.

    Check if a file exists at `filepath`.
    If not, assume that the path is relative to `base_path`.

    Args:
        filepath (str or Path): path to file
        base_path (Path): base path to resolve relative path
    """
    filepath = Path(filepath).expanduser()

    if not filepath.is_file() and not filepath.is_absolute():
        filepath = (base_path / filepath).absolute()
    return filepath



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nemo.lightning.pytorch.optim.base import LRSchedulerModule, OptimizerModule
from nemo.lightning.pytorch.optim.lr_scheduler import (
    CosineAnnealingScheduler,
    InverseSquareRootAnnealingScheduler,
    NoamAnnealingScheduler,
    NoamHoldAnnealingScheduler,
    PolynomialDecayAnnealingScheduler,
    PolynomialHoldDecayAnnealingScheduler,
    SquareAnnealingScheduler,
    SquareRootAnnealingScheduler,
    T5InverseSquareRootAnnealingScheduler,
    WarmupAnnealingScheduler,
    WarmupHoldAnnealScheduler,
    WarmupHoldPolicyScheduler,
    WarmupPolicyScheduler,
)
from nemo.lightning.pytorch.optim.megatron import MegatronOptimizerModule
from nemo.lightning.pytorch.optim.pytorch import PytorchOptimizerModule

__all__ = [
    "OptimizerModule",
    "LRSchedulerModule",
    "MegatronOptimizerModule",
    "WarmupPolicyScheduler",
    "WarmupHoldPolicyScheduler",
    "SquareAnnealingScheduler",
    "SquareRootAnnealingScheduler",
    "NoamAnnealingScheduler",
    "NoamHoldAnnealingScheduler",
    "WarmupAnnealingScheduler",
    "InverseSquareRootAnnealingScheduler",
    "T5InverseSquareRootAnnealingScheduler",
    "PolynomialDecayAnnealingScheduler",
    "PolynomialHoldDecayAnnealingScheduler",
    "CosineAnnealingScheduler",
    "PytorchOptimizerModule",
    "WarmupHoldAnnealScheduler",
]



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import torch

from nemo.collections.asr.inference.model_wrappers.ctc_inference_wrapper import CTCInferenceWrapper
from nemo.collections.asr.inference.utils.bpe_decoder import BPEDecoder
from nemo.collections.asr.inference.utils.text_segment import TextSegment, Word
from nemo.collections.asr.parts.submodules.ctc_decoding import CTCDecodingConfig


@pytest.fixture(scope="module")
def bpe_decoder():
    asr_model = CTCInferenceWrapper(
        model_name="stt_en_conformer_ctc_small",
        decoding_cfg=CTCDecodingConfig(),
        device="cuda" if torch.cuda.is_available() else "cpu",
    )
    return BPEDecoder(
        vocabulary=asr_model.get_vocabulary(),
        tokenizer=asr_model.tokenizer,
        confidence_aggregator=min,
        asr_supported_puncts=asr_model.supported_punctuation(),
        word_boundary_tolerance=0.0,  # Set to 0.0 for easy testing
        token_duration_in_secs=asr_model.get_model_stride(in_secs=True),
    )


class TestBPEDecoder:

    @pytest.mark.with_downloads
    @pytest.mark.unit
    @pytest.mark.parametrize(
        "text",
        [
            "the quick brown fox jumps over the lazy dog",
            "lorem ipsum dolor sit amet",
            "this a test sentence",
        ],
    )
    def test_group_tokens_into_words(self, bpe_decoder, text):
        ground_truth_words = text.split()
        tokens = bpe_decoder.tokenizer.text_to_ids(text)
        n_tokens = len(tokens)
        timestamps = [float(i) for i in range(n_tokens)]
        confidences = [0.1] * n_tokens

        words, need_merge = bpe_decoder.group_tokens_into_words(tokens, timestamps, confidences)
        assert len(words) == len(ground_truth_words)
        prev_word_end = -1
        for word, ground_truth_word in zip(words, ground_truth_words):
            assert isinstance(word, Word)
            assert word.text == ground_truth_word
            assert word.conf == 0.1
            assert word.end > word.start and word.start >= prev_word_end
            prev_word_end = word.end
        assert need_merge == False

    @pytest.mark.with_downloads
    @pytest.mark.unit
    @pytest.mark.parametrize(
        "text",
        [
            "the quick brown fox jumps over the lazy dog",
            "lorem ipsum dolor sit amet",
            "this a test sentence",
        ],
    )
    def test_group_tokens_into_segment(self, bpe_decoder, text):
        tokens = bpe_decoder.tokenizer.text_to_ids(text)
        n_tokens = len(tokens)
        timestamps = [float(i) for i in range(n_tokens)]
        confidences = [0.1] * n_tokens

        segment, need_merge = bpe_decoder.group_tokens_into_segment(tokens, timestamps, confidences)
        assert isinstance(segment, TextSegment)
        assert need_merge == False
        assert segment.text == text
        assert segment.start == 0.0
        assert segment.end == (n_tokens - 1) * bpe_decoder.token_duration_in_secs
        assert segment.conf == 0.1



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json

import lhotse.serialization
import pytest
from lhotse import CutSet, SupervisionSegment
from lhotse.cut import Cut
from lhotse.testing.dummies import dummy_cut, dummy_recording

from nemo.collections.common.data.lhotse import (
    NeMoSFTExample,
    SourceTargetTextExample,
    get_lhotse_dataloader_from_config,
)
from nemo.collections.common.tokenizers import SentencePieceTokenizer
from nemo.collections.common.tokenizers.sentencepiece_tokenizer import create_spt_model


@pytest.fixture
def tokenizer(tmp_path_factory):
    tmpdir = tmp_path_factory.mktemp("tok")
    text_path = tmpdir / "text.txt"
    text_path.write_text("\n".join(chr(i) for i in range(256)))
    create_spt_model(
        text_path,
        vocab_size=512,
        sample_size=-1,
        do_lower_case=False,
        output_dir=str(tmpdir),
        bos=True,
        eos=True,
        user_defined_symbols=[
            "[INST]",
            "[/INST]",
            "<<SYS>>",
            "<</SYS>>",
            "[audio]",
            "<end_of_turn>",
            "<start_of_turn>",
        ],
        remove_extra_whitespaces=True,
    )
    return SentencePieceTokenizer(str(tmpdir / "tokenizer.model"))


@pytest.fixture
def cuts_path(tmp_path_factory):
    tmp_path = tmp_path_factory.getbasetemp() / "cuts.jsonl"
    c = dummy_cut(0, duration=1.0, supervisions=[SupervisionSegment("", "", 0, 1.0, text="dummy text")])
    c.context = "dummy context"
    CutSet([c]).to_file(tmp_path)
    return tmp_path


@pytest.fixture
def src_tgt_example(tmp_path_factory):
    d = tmp_path_factory.mktemp("src_tgt")
    (d / "src.txt").write_text("an example")
    (d / "tgt.txt").write_text("elpmaxe na")
    return (d / "src.txt"), (d / "tgt.txt")


@pytest.fixture
def nemo_sft_example(tmp_path_factory):
    tmp_path = tmp_path_factory.getbasetemp() / "nemo_sft.jsonl"
    lhotse.serialization.save_to_jsonl(
        [
            {
                "system": "",
                "mask": "User",
                "dataset": "",
                "conversations": [
                    {
                        "from": "User",
                        "value": "Hi, how are you?",
                    },
                    {
                        "from": "Assistant",
                        "value": "Good day, I'm a useful assistant.",
                    },
                ],
            }
        ],
        tmp_path,
    )
    return tmp_path


@pytest.fixture
def multi_speaker_simulator_example(tmp_path_factory):
    tmp_path = tmp_path_factory.getbasetemp()

    # 1. Generate a wav file using lhotse dummy_recording with actual audio data
    wav_path = tmp_path / "wav.wav"

    # Create dummy recordings with actual audio data and save to wav files
    dummy_recording(0, duration=2.0, with_data=True).to_cut().save_audio(wav_path)

    # For type lsmix
    lsmix_manifest_path = tmp_path / "lsmix_manifest.jsonl"

    lsmix_manifest_data = [
        {
            "audio_filepath": str(wav_path),
            "session_id": "session1",
            "duration": 1.0,
            "speaker_id": "speaker1",
            "text": "session1 speaker1 dummy text",
        },
        {
            "audio_filepath": str(wav_path),
            "session_id": "session2",
            "duration": 2.0,
            "speaker_id": "speaker2",
            "text": "session2 speaker2 dummy text",
        },
    ]
    with open(lsmix_manifest_path, 'w') as f:
        for item in lsmix_manifest_data:
            f.write(json.dumps(item) + '\n')

    # For type mixture_loader
    # Generate corresponding seglst files (segment list files)
    mixture_seglst_path = tmp_path / "mixture_seglst.json"

    # Create seglst files with segment information
    mixture_seglst_data = [
        {
            "session_id": "session1",
            "start_time": "0.0",
            "end_time": "1.0",
            "speaker": "speaker1",
            "words": "session1 speaker1 dummy text",
        },
        {
            "session_id": "session1",
            "start_time": "0.5",
            "end_time": "2.0",
            "speaker": "speaker2",
            "words": "session1 speaker2 dummy text",
        },
    ]
    with open(mixture_seglst_path, 'w') as f:
        f.write(json.dumps(mixture_seglst_data))

    # Generate manifest file
    mixture_manifest_path = tmp_path / "mixture_manifest.jsonl"
    mixture_manifest_data = [
        {
            "audio_filepath": str(wav_path),
            "session_id": "session1",
            "offset": 0.0,
            "duration": 2.0,
            "seglst_filepath": str(mixture_seglst_path),
        },
    ]

    with open(mixture_manifest_path, 'w') as f:
        for item in mixture_manifest_data:
            f.write(json.dumps(item) + '\n')

    return [lsmix_manifest_path, mixture_manifest_path]


class Identity:
    def __getitem__(self, item):
        return item


def test_prompt_format_cut(cuts_path, tokenizer):
    dl = get_lhotse_dataloader_from_config(
        {
            "cuts_path": cuts_path,
            "batch_size": 1,
            "prompt_format": "llama2",
            "min_duration": 0,
            "max_duration": 10,
        },
        global_rank=0,
        world_size=1,
        dataset=Identity(),
        tokenizer=tokenizer,
    )

    batch = next(iter(dl))
    ex = batch[0]
    assert isinstance(ex, Cut)
    assert tokenizer.ids_to_text(ex.input_ids) == "[INST] dummy context [/INST] dummy text"
    assert tokenizer.ids_to_text(ex.context_ids) == "[INST] dummy context [/INST]"
    assert tokenizer.ids_to_text(ex.answer_ids) == "dummy text"


def test_prompt_format_cut_filtered_out(cuts_path, tokenizer):
    dl = get_lhotse_dataloader_from_config(
        {
            "cuts_path": cuts_path,
            "batch_size": 1,
            "prompt_format": "llama2",
            "min_duration": 0,
            "max_duration": 0.5,
        },
        global_rank=0,
        world_size=1,
        dataset=Identity(),
        tokenizer=tokenizer,
    )
    with pytest.raises(StopIteration):
        next(iter(dl))


def test_prompt_format_cut_max_tokens_has_no_filtering_effect(cuts_path, tokenizer):
    dl = get_lhotse_dataloader_from_config(
        {
            "cuts_path": cuts_path,
            "batch_size": 1,
            "prompt_format": "llama2",
            "use_multimodal_dataloading": True,
            "token_equivalent_duration": 0.1,
            "min_tokens": 1,
            "max_tokens": 2,
            "use_total_length": True,
        },
        global_rank=0,
        world_size=1,
        dataset=Identity(),
        tokenizer=tokenizer,
    )

    batch = next(iter(dl))
    ex = batch[0]
    assert isinstance(ex, Cut)


def test_prompt_format_src_tgt(src_tgt_example, tokenizer):
    dl = get_lhotse_dataloader_from_config(
        {
            "input_cfg": [
                {"type": "txt_pair", "source_paths": src_tgt_example[0], "target_paths": src_tgt_example[1]}
            ],
            "batch_size": 1,
            "force_finite": True,
            "prompt_format": "llama2",
            "use_multimodal_dataloading": True,
            "min_tokens": 1,
            "max_tokens": 50,
            "use_total_length": True,
        },
        global_rank=0,
        world_size=1,
        dataset=Identity(),
        tokenizer=tokenizer,
    )

    batch = next(iter(dl))
    ex = batch[0]
    assert isinstance(ex, SourceTargetTextExample)
    assert tokenizer.ids_to_text(ex.input_ids) == "[INST] an example [/INST] elpmaxe na"
    assert tokenizer.ids_to_text(ex.context_ids) == "[INST] an example [/INST]"
    assert tokenizer.ids_to_text(ex.answer_ids) == "elpmaxe na"


def test_prompt_format_src_tgt_filtered_out(src_tgt_example, tokenizer):
    dl = get_lhotse_dataloader_from_config(
        {
            "input_cfg": [
                {"type": "txt_pair", "source_paths": src_tgt_example[0], "target_paths": src_tgt_example[1]}
            ],
            "batch_size": 1,
            "force_finite": True,
            "prompt_format": "llama2",
            "use_multimodal_dataloading": True,
            "min_tokens": 1,
            "max_tokens": 10,
            "use_total_length": True,
        },
        global_rank=0,
        world_size=1,
        dataset=Identity(),
        tokenizer=tokenizer,
    )
    with pytest.raises(StopIteration):
        batch = next(iter(dl))


def test_prompt_format_src_tgt_2d(src_tgt_example, tokenizer):
    dl = get_lhotse_dataloader_from_config(
        {
            "input_cfg": [
                {
                    "type": "txt_pair",
                    "source_paths": src_tgt_example[0],
                    "target_paths": src_tgt_example[1],
                    "target_language": "reversed",
                }
            ],
            "batch_size": 1,
            "force_finite": True,
            "prompt_format": "t5nmt",
            "use_multimodal_dataloading": True,
            "min_tokens": 1,
            "max_tokens": 50,
            "use_total_length": False,
        },
        global_rank=0,
        world_size=1,
        dataset=Identity(),
        tokenizer=tokenizer,
    )

    batch = next(iter(dl))
    ex = batch[0]
    assert isinstance(ex, SourceTargetTextExample)
    assert tokenizer.ids_to_text(ex.input_ids) == "<reversed> an example elpmaxe na"
    assert tokenizer.ids_to_text(ex.context_ids) == "<reversed> an example"
    assert tokenizer.ids_to_text(ex.answer_ids) == "elpmaxe na"


def test_prompt_format_nemo_sft(nemo_sft_example, tokenizer):
    dl = get_lhotse_dataloader_from_config(
        {
            "input_cfg": [{"type": "nemo_sft_jsonl", "paths": nemo_sft_example}],
            "batch_size": 1,
            "force_finite": True,
            "prompt_format": "llama2",
            "use_multimodal_dataloading": True,
            "min_tokens": 1,
            "max_tokens": 100,
            "use_total_length": True,
        },
        global_rank=0,
        world_size=1,
        dataset=Identity(),
        tokenizer=tokenizer,
    )

    batch = next(iter(dl))
    ex = batch[0]
    assert isinstance(ex, NeMoSFTExample)
    assert tokenizer.ids_to_text(ex.input_ids) == "[INST] Hi, how are you? [/INST] Good day, I'm a useful assistant."
    assert tokenizer.ids_to_text(ex.context_ids) == "[INST] Hi, how are you? [/INST]"
    assert tokenizer.ids_to_text(ex.answer_ids) == "Good day, I'm a useful assistant."


def test_prompt_format_nemo_sft_filtered_out(nemo_sft_example, tokenizer):
    dl = get_lhotse_dataloader_from_config(
        {
            "input_cfg": [{"type": "nemo_sft_jsonl", "paths": nemo_sft_example}],
            "batch_size": 1,
            "force_finite": True,
            "prompt_format": "llama2",
            "use_multimodal_dataloading": True,
            "min_tokens": 1,
            "max_tokens": 5,
            "use_total_length": True,
        },
        global_rank=0,
        world_size=1,
        dataset=Identity(),
        tokenizer=tokenizer,
    )
    with pytest.raises(StopIteration):
        batch = next(iter(dl))


def test_prompt_format_multi_speaker_simulator(multi_speaker_simulator_example, tokenizer):
    lsmix_manifest_path, mixture_manifest_path = multi_speaker_simulator_example
    # For type lsmix
    dl = get_lhotse_dataloader_from_config(
        {
            "input_cfg": [
                {
                    "type": "multi_speaker_simulator",
                    "manifest_filepath": lsmix_manifest_path,
                    "simulator_type": "lsmix",
                    "num_speakers": 2,
                    "min_delay": 0.5,
                }
            ],
            "batch_size": 1,
            "min_duration": 0,
            "max_duration": 10,
        },
        global_rank=0,
        world_size=1,
        dataset=Identity(),
        tokenizer=tokenizer,
    )
    batch = next(iter(dl))
    ex = batch[0]
    assert isinstance(ex, Cut)
    assert len(ex.tracks) == 2
    texts = sorted([track.cut.supervisions[0].text for track in ex.tracks])
    assert texts == ["session1 speaker1 dummy text", "session2 speaker2 dummy text"]

    # For type mixture_loader
    dl = get_lhotse_dataloader_from_config(
        {
            "input_cfg": [
                {
                    "type": "multi_speaker_simulator",
                    "manifest_filepath": mixture_manifest_path,
                    "simulator_type": "mixture_loader",
                }
            ],
            "batch_size": 1,
            "min_duration": 0,
            "max_duration": 10,
        },
        global_rank=0,
        world_size=1,
        dataset=Identity(),
        tokenizer=tokenizer,
    )
    batch = next(iter(dl))
    ex = batch[0]
    assert isinstance(ex, Cut)
    assert len(ex.supervisions) in [1, 2]
    if len(ex.supervisions) == 1:
        assert ex.supervisions[0].text == "session1 speaker1 dummy text"
    else:
        assert ex.supervisions[0].text == "session1 speaker1 dummy text"
        assert ex.supervisions[1].text == "session1 speaker2 dummy text"



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional

import torch

from nemo.collections.asr.parts.utils.rnnt_utils import Hypothesis, NBestHypotheses
from nemo.utils.enum import PrettyStrEnum

# Constants used for hashing text sequences.
MULTIPLIER = 6364136223846793005
INCREMENT = 1
MODULUS = 2**64

# Constants used for initializing and managing beam search hypotheses.
INACTIVE_SCORE = -float("inf")  # Represents the score of inactive hypotheses.
INIT_POINTER_VALUE = -1  # Initial value for pointers in the hypothesis tree structure.
INIT_HASH_VALUE = 0  # Initial hash value for transcript hashes.
INIT_PREFIX_HASH_VALUE = 0  # Initial hash value for prefix hashes.
NON_EXISTENT_LABEL_VALUE = -1  # Placeholder value for non-existent labels in hypotheses. Needs to be negative.


def hash_text(prev_hash: torch.Tensor, add_labels: torch.Tensor) -> torch.Tensor:
    """
    Computes a new hash value by updating previous hash tensor with added labels tensor.
    Reference: https://stackoverflow.com/a/77213071

    Args:
        prev_hash (torch.Tensor): A tensor representing the previous hash value.
        add_labels (torch.Tensor): A tensor containing added labels.

    Returns:
        torch.Tensor: A tensor representing the updated hash value.
    """
    return prev_hash * MULTIPLIER + INCREMENT + add_labels


class BlankLMScoreMode(PrettyStrEnum):
    """
    Defines the strategies for handling blank token scores in a external Ngram LM
    when combined with an automatic speech recognition (ASR) model.
    """

    # No score for blank.
    NO_SCORE = "no_score"
    # Blank score for LM is set equal to blank score from ASR model; non-blank LM scores are reweighted to sum to 1.
    LM_WEIGHTED_FULL = "lm_weighted_full"


class PruningMode(PrettyStrEnum):
    """Specifies when pruning is applied external Ngram LM shallow fusion.."""

    # Hyps are pruned based on ASR probs, then rescored with LM
    EARLY = "early"
    # Hyps are scored based on combined ASR and LM probs., then pruned
    LATE = "late"


class ASRModelTypeEnum(PrettyStrEnum):
    """Specifies model type."""

    RNNT = "rnnt"
    TDT = "tdt"
    CTC = "ctc"


class BatchedBeamHyps:
    """Class to store batch of beam hypotheses (labels, time_indices, scores) for efficient batched beam decoding"""

    def __init__(
        self,
        batch_size: int,
        beam_size: int,
        init_length: int,
        blank_index: int,
        device: torch.device = None,
        float_dtype: torch.dtype = None,
        store_prefix_hashes: Optional[bool] = False,
        model_type: Optional[ASRModelTypeEnum | str] = ASRModelTypeEnum.RNNT,
    ):
        """
        Initializes the batched beam hypotheses utility for Transducer decoding (RNN-T and TDT models).
        Args:
            batch_size (int): Batch size.
            beam_size (int): Beam size.
            init_length (int): The initial maximum length of the hypotheses.
            blank_index (int): The index representing the blank token in the vocabulary.
            device (torch.device): The device on which tensors will be allocated. Defaults to None.
            float_dtype (torch.dtype): The floating-point data type. Defaults to None.
            store_prefix_hashes (bool, optional): Whether to store prefix hashes for hypotheses. Defaults to False.
            model_type: (str or ModelTypeEnum, optional): Model type, either 'rnnt', 'tdt' or 'ctc'. Defaults to 'rnnt'.
        """

        if beam_size <= 0:
            raise ValueError("Beam size must be greater than 0.")
        if batch_size <= 0:
            raise ValueError("Batch size must be greater than 0.")
        if init_length <= 0:
            raise ValueError("Initial hypothesis lengths must be greater than 0.")

        self.device = device
        self.INACTIVE_SCORE_TENSOR = torch.tensor(INACTIVE_SCORE, device=device, dtype=float_dtype)
        self.ZERO_TENSOR = torch.tensor(0, device=device, dtype=torch.long)

        self.model_type = ASRModelTypeEnum(model_type)
        self.store_prefix_hashes = store_prefix_hashes
        self._max_length = init_length
        self.beam_size = beam_size
        self.blank_index = blank_index
        self.batch_size = batch_size
        self.batch_indices = torch.arange(self.batch_size, device=device)
        self.beam_indices = torch.arange(self.beam_size, device=device)

        # Non-blank (non-blank and non-repeating for CTC) and full lengths
        self.current_lengths_nb = torch.zeros([batch_size, self.beam_size], device=device, dtype=torch.long)
        self.current_lengths_wb = torch.zeros([batch_size, self.beam_size], device=device, dtype=torch.long)

        # Initializing tree structure for hypothesis storing
        self.transcript_wb = torch.full(
            (batch_size, self.beam_size, self._max_length),
            fill_value=NON_EXISTENT_LABEL_VALUE,
            device=device,
            dtype=torch.long,
        )  # current labels
        self.transcript_wb_prev_ptr = torch.full(
            (batch_size, self.beam_size, self._max_length),
            fill_value=INIT_POINTER_VALUE,
            device=device,
            dtype=torch.long,
        )  # links to prefices

        # Initializing beam scores: Initially, only a single hypothesis is active within the beam.
        self.scores = torch.full(
            [batch_size, self.beam_size], device=device, dtype=float_dtype, fill_value=INACTIVE_SCORE
        )
        self.scores[:, 0].fill_(0.0)

        self.last_label = torch.full(
            (batch_size, self.beam_size), fill_value=NON_EXISTENT_LABEL_VALUE, device=device, dtype=torch.long
        )

        self.transcript_hash = torch.full(
            [batch_size, self.beam_size], device=device, dtype=torch.long, fill_value=INIT_HASH_VALUE
        )
        if store_prefix_hashes:
            self.transcript_prefix_hash = torch.full(
                [batch_size, self.beam_size], device=device, dtype=torch.long, fill_value=INIT_PREFIX_HASH_VALUE
            )

        if self.model_type == ASRModelTypeEnum.CTC:
            # CTC frames and tokens are aligned, so we can precompute timestamps
            self.timestamps = self._create_timestamps_tensor(self._max_length)  # timestamps
        else:
            # timestamps for transducer models
            self.timestamps = torch.zeros(
                (batch_size, self.beam_size, self._max_length), device=device, dtype=torch.long
            )  # timestamps

            # tracking last frame index and number of labels for the last frama
            self.next_timestamp = torch.zeros((batch_size, self.beam_size), device=device, dtype=torch.long)
            self.last_timestamp_lasts = torch.zeros((batch_size, self.beam_size), device=device, dtype=torch.long)

    def clear_(self):
        """
        Clears and resets the internal state of the object.
        """

        self.current_lengths_nb.fill_(0)
        self.current_lengths_wb.fill_(0)

        self.transcript_wb.fill_(NON_EXISTENT_LABEL_VALUE)
        self.transcript_wb_prev_ptr.fill_(INIT_POINTER_VALUE)

        self.scores.fill_(INACTIVE_SCORE)
        self.scores[:, 0].fill_(0.0)

        self.last_label.fill_(NON_EXISTENT_LABEL_VALUE)

        self.transcript_hash.fill_(INIT_HASH_VALUE)
        if self.store_prefix_hashes:
            self.transcript_prefix_hash.fill_(INIT_PREFIX_HASH_VALUE)

        # model specific parameters
        if self.model_type == ASRModelTypeEnum.CTC:
            self.timestamps.copy_(self._create_timestamps_tensor(self._max_length))
        else:
            self.timestamps.fill_(0)
            self.next_timestamp.fill_(0)
            self.last_timestamp_lasts.fill_(0)

    def _allocate_more(self):
        """
        Dynamically allocates more memory for the internal buffers.
        This method doubles the size of the following tensors: `transcript_wb`, `transcript_wb_prev_ptr`.
        """
        self.transcript_wb = torch.cat(
            (self.transcript_wb, torch.full_like(self.transcript_wb, fill_value=NON_EXISTENT_LABEL_VALUE)), dim=-1
        )
        self.transcript_wb_prev_ptr = torch.cat(
            (self.transcript_wb_prev_ptr, torch.full_like(self.transcript_wb_prev_ptr, fill_value=INIT_POINTER_VALUE)),
            dim=-1,
        )
        if self.model_type == ASRModelTypeEnum.CTC:
            self.timestamps = self._create_timestamps_tensor(2 * self._max_length)
        else:
            self.timestamps = torch.cat((self.timestamps, torch.zeros_like(self.timestamps)), dim=-1)

        self._max_length *= 2

    def add_results_(
        self,
        next_indices: torch.Tensor,
        next_labels: torch.Tensor,
        next_hyps_prob: torch.Tensor,
        next_label_durations: Optional[torch.Tensor] = None,
    ):
        """
        Updates batch of beam hypotheses with labels. If the maximum allowed length
        is exceeded, underlying memory is doubled.
        Args:
            next_indices (torch.Tensor): Indices of the hypotheses to be updated.
            next_labels (torch.Tensor): Labels corresponding to the next step in the beam search.
            next_hyps_prob (torch.Tensor): Probabilities of the next hypotheses.
            next_label_durations (torch.Tensor, optional): Durations associated with the next labels. Required when `model_type='tdt'`.
        """

        if self.model_type == ASRModelTypeEnum.TDT and next_label_durations is None:
            raise ValueError("`next_label_durations` is required when model type is TDT.")

        if (self.current_lengths_wb + 1).max() >= self._max_length:
            self._allocate_more()

        self.add_results_no_checks_(
            next_indices=next_indices,
            next_labels=next_labels,
            next_hyps_prob=next_hyps_prob,
            next_label_durations=next_label_durations,
        )

    def add_results_no_checks_(
        self,
        next_indices: torch.Tensor,
        next_labels: torch.Tensor,
        next_hyps_prob: torch.Tensor,
        next_label_durations: Optional[torch.Tensor] = None,
    ):
        """
        Updates batch of beam hypotheses with labels.
        Args:
            next_indices (torch.Tensor): Indices of the hypotheses to be updated.
            next_labels (torch.Tensor): Labels corresponding to the next step in the beam search.
            next_hyps_prob (torch.Tensor): Probabilities of the next hypotheses.
            next_label_durations (torch.Tensor, optional): Durations associated with the next labels. Required when `model_type='tdt'`.
        """
        if self.model_type == ASRModelTypeEnum.TDT and next_label_durations is None:
            raise ValueError("`next_label_durations` is required when model type is TDT.")

        last_labels = torch.gather(self.last_label, dim=-1, index=next_indices)
        self.transcript_wb.scatter_(dim=-1, index=self.current_lengths_wb.unsqueeze(-1), src=next_labels.unsqueeze(-1))
        self.transcript_wb_prev_ptr.scatter_(
            dim=-1, index=self.current_lengths_wb.unsqueeze(-1), src=next_indices.unsqueeze(-1)
        )

        is_extended = next_labels >= 0
        extended_with_blank = next_labels == self.blank_index
        extended_with_label = (is_extended) & (~extended_with_blank)
        if self.model_type == ASRModelTypeEnum.CTC:
            # for CTC last non-blank and non-repeated label
            extended_with_label = (extended_with_label) & (next_labels != last_labels)  # non-repeated non-blank label

        if self.model_type == ASRModelTypeEnum.RNNT:
            timesteps = torch.gather(self.next_timestamp, dim=-1, index=next_indices)
            self.timestamps.scatter_(
                dim=-1,
                index=self.current_lengths_wb.unsqueeze(-1),
                src=(timesteps + extended_with_blank).unsqueeze(-1),
            )
            self.next_timestamp.copy_(timesteps + extended_with_blank)
            torch.where(
                extended_with_blank,
                self.ZERO_TENSOR,
                torch.gather(self.last_timestamp_lasts, dim=-1, index=next_indices) + extended_with_label,
                out=self.last_timestamp_lasts,
            )
        elif self.model_type == ASRModelTypeEnum.TDT:
            timesteps = torch.gather(self.next_timestamp, dim=-1, index=next_indices)
            next_label_durations = torch.where(is_extended, next_label_durations, 0)
            self.timestamps.scatter_(
                dim=-1,
                index=self.current_lengths_wb.unsqueeze(-1),
                src=(timesteps + next_label_durations).unsqueeze(-1),
            )
            torch.where(is_extended, timesteps + next_label_durations, timesteps, out=self.next_timestamp)
            torch.where(
                is_extended & (next_label_durations > 0),
                self.ZERO_TENSOR,
                torch.gather(self.last_timestamp_lasts, dim=-1, index=next_indices) + extended_with_label,
                out=self.last_timestamp_lasts,
            )

        self.current_lengths_nb.copy_(
            torch.gather(self.current_lengths_nb, dim=-1, index=next_indices) + extended_with_label
        )
        torch.add(self.current_lengths_wb, 1, out=self.current_lengths_wb)
        self.scores.copy_(next_hyps_prob)

        prev_transcript_hash = torch.gather(self.transcript_hash, dim=-1, index=next_indices)
        # update hashes and prefix hashes
        torch.where(
            extended_with_label,
            hash_text(prev_transcript_hash, next_labels),
            prev_transcript_hash,
            out=self.transcript_hash,
        )

        if self.model_type == ASRModelTypeEnum.CTC:
            # track last label
            torch.where(is_extended, next_labels, last_labels, out=self.last_label)
        else:
            # track last non-blank label
            torch.where(extended_with_label, next_labels, last_labels, out=self.last_label)

        # store prefix hashes for batched maes
        if self.store_prefix_hashes:
            prev_transcript_prefix_hash = torch.gather(self.transcript_prefix_hash, dim=-1, index=next_indices)
            torch.where(
                extended_with_label, prev_transcript_hash, prev_transcript_prefix_hash, out=self.transcript_prefix_hash
            )

    def recombine_hyps_(self):
        """
        Recombines hypotheses in the beam search by merging equivalent hypotheses and updating their scores.
        This method identifies hypotheses that are equivalent based on their transcript hash, last label,
        and current lengths. It then merges these equivalent hypotheses by computing a new score using
        log-sum-exp over their scores and updates the scores tensor accordingly.
        Returns:
            Note: The method modifies the `self.scores` tensor in place to reflect the recombined hypotheses.
        """

        if self.beam_size <= 1:
            return

        hyps_equal = (
            (self.transcript_hash[:, :, None] == self.transcript_hash[:, None, :])
            & (self.last_label[:, :, None] == self.last_label[:, None, :])
            & (self.current_lengths_nb[:, :, None] == self.current_lengths_nb[:, None, :])
        )

        if self.model_type == ASRModelTypeEnum.TDT:
            hyps_equal &= self.next_timestamp[:, :, None] == self.next_timestamp[:, None, :]

        scores_matrix = torch.where(
            hyps_equal,
            self.scores[:, None, :].expand(self.batch_size, self.beam_size, self.beam_size),
            self.INACTIVE_SCORE_TENSOR,
        )
        scores_argmax = scores_matrix.argmax(-1, keepdim=False)
        scores_to_keep = (
            torch.arange(self.beam_size, device=scores_argmax.device, dtype=torch.long)[None, :] == scores_argmax
        )
        if self.model_type == ASRModelTypeEnum.CTC:
            new_scores = torch.max(scores_matrix, dim=-1, keepdim=False).values
        else:
            new_scores = torch.logsumexp(scores_matrix, dim=-1, keepdim=False)
        torch.where(scores_to_keep, new_scores.to(self.scores.dtype), self.INACTIVE_SCORE_TENSOR, out=self.scores)

    def remove_duplicates(self, labels: torch.Tensor, total_logps: torch.Tensor):
        """
        Removes duplicate hypotheses that may arise after updating beam hypotheses with labels during the beam search process.
        Args:
            labels (torch.Tensor): A tensor containing the labels for the current beam
                search step. Shape: [batch_size, beam_size, ...].
            total_logps (torch.Tensor): A tensor containing the total log probabilities
                for the current beam search step. Shape: [batch_size, beam_size, ...].
        Returns:
            torch.Tensor: Updated total log probabilities with duplicates removed.
                Shape: [batch_size, beam_size, ...].
        """

        if self.beam_size <= 1:
            return total_logps

        # updating hashes for label expansions
        non_blank_mask = labels != self.blank_index
        expansion_hashes = hash_text(self.transcript_hash.unsqueeze(-1), labels)
        expansion_hashes = torch.where(non_blank_mask, expansion_hashes, self.transcript_hash.unsqueeze(-1)).view(
            self.batch_size, -1
        )

        # masking inactive hypotheses
        inactive_hyps_mask = self.scores != INACTIVE_SCORE
        masked_hashes = torch.where(inactive_hyps_mask, self.transcript_hash, -1)

        init_expansions_equal = (expansion_hashes[:, :, None] == masked_hashes[:, None, :]).any(dim=-1)

        init_expansions_equal = torch.logical_and(non_blank_mask.view(self.batch_size, -1), init_expansions_equal)
        expansions_equal = expansion_hashes[:, :, None] == expansion_hashes[:, None, :]
        expansion_scores = total_logps.view(self.batch_size, -1)
        expansion_scores = torch.where(init_expansions_equal, INACTIVE_SCORE, expansion_scores)
        expansion_scores = expansion_scores[:, None, :].expand(expansions_equal.shape)

        expansion_scores = torch.where(expansions_equal, expansion_scores, INACTIVE_SCORE)
        expansion_scores, expansion_scores_argmax = expansion_scores.max(dim=-1)

        scores_range = torch.arange(
            expansion_scores_argmax.shape[-1], device=expansion_scores_argmax.device, dtype=torch.long
        )
        scores_to_keep = scores_range[None, :] == expansion_scores_argmax
        total_logps = torch.where(scores_to_keep, expansion_scores, INACTIVE_SCORE).view(
            self.batch_size, self.beam_size, -1
        )

        return total_logps

    def recombine_prefixes(self, label_logps: torch.Tensor, active_mask: torch.Tensor):
        """
        Recombines prefixes (prefix search) in the beam search process by updating scores for hypotheses
        that share common prefixes.
        Args:
            label_logps (torch.Tensor): A tensor of shape (batch_size, beam_size, vocab_size)
                containing the log probabilities of the labels for each beam.
            active_mask (torch.Tensor): A boolean tensor of shape (batch_size, beam_size)
                indicating which beams are active.
        """

        if self.beam_size <= 1:
            return

        # if hypotheses are empty skip
        if (self.current_lengths_wb == 0).any():
            return

        # mask prefix hashes if hypotheses of the beam do not have prefixes (e.g. no non-blank labels were appended)
        prefix_hashes = torch.where(self.current_lengths_nb == 0, -2, self.transcript_prefix_hash)

        prefix_equal = self.transcript_hash[:, None, :] == prefix_hashes[:, :, None]

        last_labels = torch.where(self.last_label == NON_EXISTENT_LABEL_VALUE, self.blank_index, self.last_label)
        prefix_labels = last_labels.unsqueeze(1).repeat((1, self.beam_size, 1))
        prefix_scores = self.scores.unsqueeze(1).repeat((1, self.beam_size, 1))

        prefix_label_logps = torch.gather(label_logps, dim=-1, index=prefix_labels)
        prefix_label_logps = prefix_scores + prefix_label_logps.transpose(dim0=-1, dim1=-2)
        prefix_label_logps = torch.where(prefix_equal, prefix_label_logps, INACTIVE_SCORE)
        prefix_label_logps = torch.logsumexp(prefix_label_logps, dim=-1)

        to_update_mask = torch.logical_and(active_mask, self.scores != INACTIVE_SCORE)
        self.scores = torch.where(to_update_mask, torch.logaddexp(self.scores, prefix_label_logps), self.scores)

    def to_hyps_list(self, score_norm: bool = True) -> list[Hypothesis]:
        """
        Converts the batched beam search results into a list of signle best hypotheses for each batch.
        Args:
            score_norm (bool):  If True, normalize the scores before sorting. Defaults to True.
        Returns:
            list[Hypothesis]: A list where each element corresponds to a batch and contains
            best hypothesis.
        """
        self.flatten_sort_(score_norm)

        scores = self.scores[self.batch_indices, 0].tolist()

        max_idx = self.current_lengths_wb.max() - 1
        timestamps = self.timestamps[..., 0, : max_idx + 1]
        transcripts = self.transcript_wb[..., 0, : max_idx + 1]
        hypotheses = [
            Hypothesis(
                score=scores[batch_idx],
                y_sequence=transcripts[batch_idx][mask := self._create_transcripts_mask(transcripts[batch_idx])]
                .cpu()
                .detach()
                .numpy(),
                timestamp=timestamps[batch_idx][mask].cpu().detach().numpy(),
                alignments=None,
                dec_state=None,
            )
            for batch_idx in range(self.batch_size)
        ]
        return hypotheses

    def to_nbest_hyps_list(self, score_norm: bool = True) -> list[NBestHypotheses]:
        """
        Converts the batched beam search results into a list of N-best hypotheses for each batch.
        Args:
            score_norm (bool, optional): If True, normalize the scores before sorting. Defaults to True.
        Returns:
            list[NBestHypotheses]: A list where each element corresponds to a batch and contains
            N-best hypotheses.
        """

        self.flatten_sort_(score_norm)

        scores = self.scores.tolist()

        max_idx = self.current_lengths_wb.max() - 1
        transcripts = self.transcript_wb[..., : max_idx + 1]
        timestamps = self.timestamps[..., : max_idx + 1]
        hypotheses = [
            NBestHypotheses(
                [
                    Hypothesis(
                        score=scores[batch_idx][beam_idx],
                        y_sequence=transcripts[batch_idx][beam_idx][
                            mask := self._create_transcripts_mask(transcripts[batch_idx][beam_idx])
                        ]
                        .cpu()
                        .detach()
                        .numpy(),
                        timestamp=timestamps[batch_idx][beam_idx][mask].cpu().detach().numpy(),
                        alignments=None,
                        dec_state=None,
                    )
                    for beam_idx in range(self.beam_size)
                    if scores[batch_idx][beam_idx] > INACTIVE_SCORE
                ]
            )
            for batch_idx in range(self.batch_size)
        ]
        return hypotheses

    def flatten_sort_(self, score_norm: bool = True):
        """
        Sorts and flattens the tree structure of hypotheses in a batched beam search decoding process.
        Args:
            score_norm (bool, optional): If True, normalizes the scores by dividing
                them by the current lengths of the hypotheses plus one. Defaults to True.
        This method performs the following steps:
        1. Normalizes the scores if `score_norm` is True.
        2. Sorts the normalized scores in descending order and retrieves the corresponding indices.
        3. Iteratively reconstructs the tokens and timestamps for each hypothesis in reverse order.
        4. Updates the internal state of the object, including transcripts, timestamps, scores,
           lengths, labels, and other metadata, based on the sorted order.
        """

        # add one for consistency with non-batched decodings, that use SOS.
        normalized_scores = (
            self.scores / (self.current_lengths_nb.to(self.scores.dtype) + 1) if score_norm else self.scores
        )
        normalized_scores, indices = torch.sort(normalized_scores, dim=-1, descending=True)

        max_idx = self.current_lengths_wb.max() - 1
        ptrs = indices

        for idx in range(max_idx, -1, -1):
            self.transcript_wb[..., idx].copy_(self.transcript_wb[self.batch_indices.unsqueeze(-1), ptrs, idx])
            if self.model_type == ASRModelTypeEnum.TDT or self.model_type == ASRModelTypeEnum.RNNT:
                self.timestamps[..., idx].copy_(self.timestamps[self.batch_indices.unsqueeze(-1), ptrs, idx])
            ptrs = self.transcript_wb_prev_ptr[self.batch_indices.unsqueeze(-1), ptrs, idx]
        self.transcript_wb_prev_ptr[..., : max_idx + 1].copy_(self.beam_indices.unsqueeze(0).unsqueeze(-1))

        self.scores.copy_(torch.gather(self.scores, dim=-1, index=indices))
        self.current_lengths_nb.copy_(torch.gather(self.current_lengths_nb, dim=-1, index=indices))
        self.current_lengths_wb.copy_(torch.gather(self.current_lengths_wb, dim=-1, index=indices))

        self.last_label.copy_(torch.gather(self.last_label, dim=-1, index=indices))

        if self.model_type == ASRModelTypeEnum.TDT or self.model_type == ASRModelTypeEnum.RNNT:
            self.next_timestamp.copy_(torch.gather(self.next_timestamp, dim=-1, index=indices))
            self.last_timestamp_lasts.copy_(torch.gather(self.last_timestamp_lasts, dim=-1, index=indices))

        self.transcript_hash.copy_(torch.gather(self.transcript_hash, dim=-1, index=indices))
        if self.store_prefix_hashes:
            self.transcript_prefix_hash.copy_(torch.gather(self.transcript_prefix_hash, dim=-1, index=indices))

    def _create_fold_consecutive_mask(self, transcript):
        """
        Creates a mask to filter consecutive duplicates, blanks, and invalid tokens in a transcript.
        Args:
            transcript (torch.Tensor): 1D tensor of token sequence.
        Returns:
            torch.Tensor: Boolean mask indicating valid tokens.
        """
        device = transcript.device
        mask = (
            (transcript >= 0)
            & torch.cat([torch.tensor([True], device=device), transcript[1:] != transcript[:-1]])
            & (transcript != self.blank_index)
        )

        return mask

    def _create_timestamps_tensor(self, max_time):
        """
        Generates a tensor of timestamps.

        In CTC, labels align with input frames, allowing timestamps to be precomputed.

        Args:
            max_time (int): The maximum number of time steps (frames) to include in the tensor.

        Returns:
            torch.Tensor: A tensor of shape (batch_size, beam_size, max_time) containing
            sequential timestamps for each batch and beam.
        """
        return torch.arange(max_time, device=self.device, dtype=torch.long)[None, None, :].repeat(
            self.batch_size, self.beam_size, 1
        )

    def _create_transcripts_mask(self, transcripts: torch.Tensor):
        """
        Processes the transcripts.
        For RNN-T and TDT removes blanks.
        For CTC removes remove consecutive duplicates and blanks.
        Args:
            transcripts (torch.Tensor): 1D tensor of token sequence.
        Returns:
            torch.Tensor: Binary mask indicating valid tokens.
        """
        if self.model_type == ASRModelTypeEnum.CTC:
            return self._create_fold_consecutive_mask(transcripts)
        else:
            return (transcripts >= 0) & (transcripts != self.blank_index)



# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse

import numpy as np
import pandas as pd


def construct_negatives(input_file, output_file, num_passages, num_negatives):
    qrels = pd.read_csv(input_file, delimiter="\t", header=None)
    with open(output_file, "w") as f:
        for i in range(len(qrels)):
            query_id, rel_passage_id = qrels[0][i], qrels[2][i]
            negatives = np.random.randint(num_passages, size=num_negatives)
            output_ids = [query_id, rel_passage_id] + negatives.tolist()
            output_str = [str(id_) for id_ in output_ids]
            print("\t".join(output_str), file=f)


def main():
    parser = argparse.ArgumentParser(description="Negative passages construction")
    parser.add_argument("--data", type=str, default="msmarco_dataset", help="path to folder with data")
    parser.add_argument("--num_passages", type=int, default=8841823, help="total number of passages")
    parser.add_argument("--num_negatives", type=int, default=10, help="number of negatives per positive")
    args = parser.parse_args()

    for mode in ["train", "dev"]:
        construct_negatives(
            input_file=f"{args.data}/qrels.{mode}.tsv",
            output_file=f"{args.data}/query2passages.{mode}.tsv",
            num_passages=args.num_passages,
            num_negatives=args.num_negatives,
        )


if __name__ == '__main__':
    main()



# ******************************************************************************
# Copyright (C) 2024 NVIDIA CORPORATION
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ******************************************************************************
"""The continuous image tokenizer with VAE or AE formulation for 2D data."""

from collections import OrderedDict, namedtuple

import torch
from torch import nn

from nemo.collections.common.video_tokenizers.modules import ContinuousFormulation, DecoderType, EncoderType

NetworkEval = namedtuple("NetworkEval", ["reconstructions", "posteriors", "latent"])


class ContinuousImageTokenizer(nn.Module):
    def __init__(self, z_channels: int, z_factor: int, latent_channels: int, **kwargs) -> None:
        super().__init__()
        self.name = kwargs.get("name", "ContinuousImageTokenizer")
        self.latent_channels = latent_channels

        encoder_name = kwargs.get("encoder", EncoderType.Default.name)
        self.encoder = EncoderType[encoder_name].value(z_channels=z_factor * z_channels, **kwargs)

        decoder_name = kwargs.get("decoder", DecoderType.Default.name)
        self.decoder = DecoderType[decoder_name].value(z_channels=z_channels, **kwargs)

        self.quant_conv = torch.nn.Conv2d(z_factor * z_channels, z_factor * latent_channels, 1)
        self.post_quant_conv = torch.nn.Conv2d(latent_channels, z_channels, 1)

        formulation_name = kwargs.get("formulation", ContinuousFormulation.AE.name)
        self.distribution = ContinuousFormulation[formulation_name].value()

        num_parameters = sum(param.numel() for param in self.parameters())

    def encoder_jit(self):
        return nn.Sequential(
            OrderedDict(
                [
                    ("encoder", self.encoder),
                    ("quant_conv", self.quant_conv),
                    ("distribution", self.distribution),
                ]
            )
        )

    def decoder_jit(self):
        return nn.Sequential(
            OrderedDict(
                [
                    ("post_quant_conv", self.post_quant_conv),
                    ("decoder", self.decoder),
                ]
            )
        )

    def last_decoder_layer(self):
        return self.decoder.conv_out

    def encode(self, x):
        h = self.encoder(x)
        moments = self.quant_conv(h)
        return self.distribution(moments)

    def decode(self, z):
        z = self.post_quant_conv(z)
        dec = self.decoder(z)
        return dec

    def forward(self, input) -> dict[str, torch.Tensor] | NetworkEval:
        latent, posteriors = self.encode(input)
        dec = self.decode(latent)
        if self.training:
            return dict(reconstructions=dec, posteriors=posteriors, latent=latent)
        return NetworkEval(reconstructions=dec, posteriors=posteriors, latent=latent)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse

from nemo_run.config import get_nemorun_home

from .utils import DEFAULT_NEMO_HOME


def parse_cli_args():
    """
    Command line arguments correspong to Slurm cluster and NeMo2.0 for running pre-training and
    fine-tuning experiments.
    """
    parser = argparse.ArgumentParser(description="NeMo2.0 Performance Pretraining and Fine-Tuning")

    parser.add_argument(
        "-a",
        "--account",
        type=str,
        help="Slurm account to use for experiment",
        required=True,
    )
    parser.add_argument(
        "-p",
        "--partition",
        type=str,
        help="Slurm partition to use for experiment",
        required=True,
    )
    parser.add_argument(
        "-g",
        "--gpu",
        type=str,
        choices=["h100", "b200", "gb200"],
        help="Target gpu type.",
        required=True,
    )
    parser.add_argument(
        "-l",
        "--log_dir",
        type=str,
        help=f"Directory for logging experiment results. Defaults to {get_nemorun_home()}",
        required=False,
        default=get_nemorun_home(),
    )
    parser.add_argument(
        "-t",
        "--time_limit",
        type=str,
        help="Maximum time limit to run experiment for. Defaults to 30 minutes (format- 'HH:MM:SS')",
        required=False,
        default="00:30:00",
    )

    parser.add_argument(
        "--additional_slurm_params",
        type=str,
        help="Additional SLURM parameters as key=value pairs. "
        "Use semicolons (;) to separate parameters when values contain commas. "
        "Examples: 'nodelist=node001,node002;constraint=gpu' or 'reservation=my_res;exclusive'",
        required=False,
        default=None,
    )
    container_img_msg = [
        "NeMo container to use for experiment. Defaults to latest dev container- 'nvcr.io/nvidia/nemo:dev'",
        "Make sure your NGC credentials are accessible in your environment.",
    ]
    parser.add_argument(
        "-i",
        "--container_image",
        type=str,
        help=" ".join(container_img_msg),
        required=False,
        default="nvcr.io/nvidia/nemo:dev",
    )
    parser.add_argument(
        "-c",
        "--compute_dtype",
        type=str,
        choices=["bf16", "fp8"],
        help="Compute precision. Options- bf16 or fp8. Defaults to bf16",
        required=False,
        default="bf16",
    )
    fp8_recipe_msg = (
        "FP8 recipe. Options- ds (per-tensor delayed scaling), cs (per-tensor current scaling), "
        "mxfp8, ss (subchannel scaling). Defaults to ds"
    )
    parser.add_argument(
        "-fr",
        "--fp8_recipe",
        type=str,
        choices=["ds", "cs", "mxfp8", "ss"],
        help=fp8_recipe_msg,
        required=False,
        default="ds",
    )
    parser.add_argument(
        "-en",
        "--enable_nsys",
        help="Enable Nsys profiling. Diabled by default",
        action="store_true",
    )
    parser.add_argument(
        "-em",
        "--enable_memory_profile",
        help="Enable memory usage profiling. Diabled by default",
        action="store_true",
    )
    parser.add_argument(
        "-mp",
        "--memory_profile_out_path",
        type=str,
        help="Path to the output file of memory profiling",
        required=False,
        default=None,
    )
    parser.add_argument(
        "-tb",
        "--tensorboard",
        help="Enable tensorboard logging. Disabled by default",
        action="store_true",
    )
    parser.add_argument(
        "-wd",
        "--wandb",
        help="Enable wandb logging. Disabled by default",
        action="store_true",
    )
    parser.add_argument(
        "-wdk",
        "--wandb_key",
        type=str,
        help="wandb key. Needed for wandb logger projetion to server",
        required=False,
        default=None,
    )
    parser.add_argument(
        "-wdp",
        "--wandb_prj_name",
        type=str,
        help="wandb project name",
        required=False,
        default=None,
    )
    parser.add_argument(
        "-wdj",
        "--wandb_job_name",
        type=str,
        help="wandb job name",
        required=False,
        default=None,
    )
    parser.add_argument(
        "-f",
        "--finetuning",
        choices=["sft", "lora"],
        help="Finetuning scheme to use. Defaults to 'lora'",
        default='lora',
    )
    parser.add_argument(
        "-hf",
        "--hf_token",
        type=str,
        help="HuggingFace token. Defaults to None. Required for accessing tokenizers and checkpoints.",
        default=None,
    )
    nemo_home_msg = [
        "Sets env var `NEMO_HOME` (on compute node using sbatch script)- directory where NeMo searches",
        "for models and datasets. This saves a lot of time (especially for bigger models) if checkpoints already",
        f"exist here. Missing files will be downloaded here from HuggingFace. Defaults to {DEFAULT_NEMO_HOME}",
    ]
    parser.add_argument(
        "-nh",
        "--nemo_home",
        type=str,
        help=" ".join(nemo_home_msg),
        default=DEFAULT_NEMO_HOME,
    )
    parser.add_argument(
        "-d",
        "--dryrun",
        help="If true, prints sbatch script to terminal without launching experiment.",
        required=False,
        action="store_true",
    )
    parser.add_argument(
        "-tp",
        "--tensor_parallel_size",
        type=int,
        help="Intra-layer model parallelism. Splits tensors across GPU ranks.",
        required=False,
        default=None,
    )
    parser.add_argument(
        "-pp",
        "--pipeline_parallel_size",
        type=int,
        help="Inter-layer model parallelism. Splits transformer layers across GPU ranks.",
        required=False,
        default=None,
    )
    parser.add_argument(
        "-cp",
        "--context_parallel_size",
        type=int,
        help="Splits network input along sequence dimension across GPU ranks.",
        required=False,
        default=None,
    )
    parser.add_argument(
        "-vp",
        "--virtual_pipeline_parallel_size",
        type=int,
        help="Number of virtual blocks per pipeline model parallel rank is the virtual model parallel size.",
        required=False,
        default=None,
    )
    parser.add_argument(
        "-ep",
        "--expert_parallel_size",
        type=int,
        help="Distributes Moe Experts across sub data parallel dimension.",
        required=False,
        default=None,
    )
    parser.add_argument(
        "-et",
        "--expert_tensor_parallel_size",
        type=lambda x: int(x) if x is not None else None,
        nargs="?",
        const=None,
        help="Intra-layer tensor model parallelsm for expert layer. Splits tensors across GPU ranks.\
            Use -et/--expert_tensor_parallel_size <space> for None or -et/--expert_tensor_parallel_size <int>",
        required=False,
        default=None,
    )
    parser.add_argument(
        "-mb",
        "--micro_batch_size",
        type=int,
        required=False,
        default=None,
    )
    parser.add_argument(
        "-gb",
        "--global_batch_size",
        type=int,
        required=False,
        default=None,
    )
    parser.add_argument(
        "-ng",
        "--num_gpus",
        type=int,
        help="Number of gpus.",
        required=False,
        default=None,
    )
    parser.add_argument(
        "-gn",
        "--gpus_per_node",
        type=int,
        help="Number of gpus per node. Defaults to 8",
        required=False,
        default=8,
    )
    parser.add_argument(
        "-ms",
        "--max_steps",
        type=int,
        help="Number of train steps. Defaults to 100",
        required=False,
        default=100,
    )

    def bool_arg(arg):
        if arg.lower() in ['true', '1', 't', 'yes', 'y']:
            return True
        elif arg.lower() in ['false', '0', 'f', 'no', 'n']:
            return False
        else:
            raise ValueError(f"Invalid value for boolean argument: {arg}")

    parser.add_argument(
        "-cg",
        "--cuda_graphs",
        help="Enable CUDA graphs. Disabled by default",
        type=bool_arg,
        required=False,
        default=None,  # NOTE: DO NOT SET DEFAULT TO FALSE, IT WILL BE OVERRIDDEN BY THE RECOMMENDED MODEL CONFIGS
    )
    parser.add_argument(
        "-fsdp",
        "--use_mcore_fsdp",
        help="Enable Megatron Core (Mcore) FSDP. Disabled by default",
        type=bool_arg,
        required=False,
        default=None,
    )
    parser.add_argument(
        "-fsdp_db",
        "--use_fsdp_double_buffer",
        help="Enable FSDP double buffer. Disabled by default",
        type=bool_arg,
        required=False,
        default=None,
    )
    parser.add_argument(
        "-ubr",
        "--use_user_buffer_registration",
        help="Enable user buffer registration. Disabled by default",
        type=bool_arg,
        required=False,
        default=None,
    )
    parser.add_argument(
        "-sharp",
        "--use_sharp",
        help="Enable sharp. Disabled by default",
        type=bool_arg,
        required=False,
        default=None,
    )
    parser.add_argument(
        "-rl",
        "--recompute_layers",
        type=int,
        help="Number of Transformer layers to recompute, where all the intermediate "
        "activations of a Transformer layer are computed. Defaults to None",
        required=False,
        default=None,
    )
    parser.add_argument(
        "-ol",
        "--activation_offload_layers",
        type=int,
        help="Number of Transformer layers to offload to the CPU memory. Defaults to None",
        required=False,
        default=None,
    )
    parser.add_argument(
        "--nccl_communicator_config_path",
        type=str,
        help="Path to NCCL communicator config yaml file",
        required=False,
        default=None,
    )

    def list_of_strings(arg):
        return arg.split(',')

    parser.add_argument(
        "-rm",
        "--recompute_modules",
        nargs="*",
        const=None,
        type=str,
        help="List of modules to perform selective activation recompute. "
        "Users can provide 0 or any number of arguments. Defaults to None",
        required=False,
        default=None,
    )
    parser.add_argument(
        "-cm",
        "--custom_mounts",
        type=list_of_strings,
        help="Comma separated string of mounts",
        required=False,
        default=[],
    )
    parser.add_argument(
        "--use_hf_tokenizer",
        help="Use HuggingFace tokenizer. Disabled by default. Null tokenizer will be used if not provided.",
        action="store_true",
        required=False,
    )
    parser.add_argument(
        "-dcdfr",
        "--dump_config_diff_from_base_recipe",
        help="Dump the config diff from the base recipe. Defaults to False",
        action="store_true",
        required=False,
        default=False,
    )
    parser.add_argument(
        "--keep_fsdp_fp8_transpose_cache",
        help="Keep FSDP FP8 transpose cache. Disabled by default",
        type=bool_arg,
        required=False,
        default=None,
    )
    parser.add_argument(
        "-vb",
        "--enable_vboost",
        help="Enable VBoost which steers more power towards tensor cores. Disabled by default",
        type=bool_arg,
        required=False,
        default=None,
    )
    parser.add_argument(
        "--use_te_op_fuser",
        help="Enable Transformer Engine's operation fuser. This feature is experimental and disabled by default",
        type=bool_arg,
        required=False,
        default=None,
    )
    parser.add_argument(
        "--use_te_act_func",
        help="Use TE activation function for the MLP part.",
        type=bool_arg,
        required=False,
        default=None,
    )
    parser.add_argument(
        "--act_func_fp8_input_store",
        help="Store input of activation function in FP8 (tensorwise recipe). Disabled by default",
        type=bool_arg,
        required=False,
        default=False,
    )
    parser.add_argument(
        "--detach",
        help="Detach from experiment. Default is True. Unset to keep the process running.",
        type=bool_arg,
        required=False,
        default=True,
    )

    return parser


def parse_additional_slurm_params(params_str):
    """
    Parse additional SLURM parameters from a string of key=value pairs.

    This function handles different separator formats:
    1. Semicolon-separated: "key1=value1;key2=value2" (recommended for multiple parameters)
    2. Space-separated: "key1=value1 key2=value2"
    3. Single parameter: "key1=value1,value2" (no separators = single parameter)

    Args:
        params_str (str): String with parameters

    Returns:
        dict: Dictionary of parameters, or None if params_str is None/empty

    Example:
        parse_additional_slurm_params("nodelist=node001,node002")
        returns {"nodelist": "node001,node002"}

        parse_additional_slurm_params("nodelist=node001,node002;constraint=gpu")
        returns {"nodelist": "node001,node002", "constraint": "gpu"}

        parse_additional_slurm_params("reservation=my_res;constraint=gpu")
        returns {"reservation": "my_res", "constraint": "gpu"}
    """
    if not params_str:
        return None

    params = {}

    # Try semicolon separation first (most reliable for complex values)
    if ';' in params_str:
        parts = params_str.split(';')
    # Try space separation next
    elif ' ' in params_str:
        parts = params_str.split()
    # No separators found - treat as single parameter
    else:
        parts = [params_str]

    for part in parts:
        part = part.strip()
        if not part:
            continue

        if '=' in part:
            key, value = part.split('=', 1)
            params[key.strip()] = value.strip()
        else:
            # Boolean flag (no value)
            params[part] = True

    return params if params else None



# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from pathlib import Path

import torch
import torch.nn.functional as F
from hydra.utils import instantiate
from lightning.pytorch.loggers.wandb import WandbLogger
from omegaconf import DictConfig, OmegaConf, open_dict

from nemo.collections.tts.losses.hifigan_losses import DiscriminatorLoss, FeatureMatchingLoss, GeneratorLoss
from nemo.collections.tts.models.base import Vocoder
from nemo.collections.tts.modules.hifigan_modules import MultiPeriodDiscriminator, MultiScaleDiscriminator
from nemo.collections.tts.parts.utils.callbacks import LoggingCallback
from nemo.collections.tts.parts.utils.helpers import get_batch_size, get_num_workers, plot_spectrogram_to_numpy
from nemo.core.classes import Exportable
from nemo.core.classes.common import PretrainedModelInfo, typecheck
from nemo.core.neural_types.elements import AudioSignal, MelSpectrogramType
from nemo.core.neural_types.neural_type import NeuralType
from nemo.core.optim.lr_scheduler import compute_max_steps, prepare_lr_scheduler
from nemo.utils import logging, model_utils

HAVE_WANDB = True
try:
    import wandb
except ModuleNotFoundError:
    HAVE_WANDB = False


class HifiGanModel(Vocoder, Exportable):
    """
    HiFi-GAN model (https://arxiv.org/abs/2010.05646) that is used to generate audio from mel spectrogram.
    """

    def __init__(self, cfg: DictConfig, trainer: 'Trainer' = None):
        # Convert to Hydra 1.0 compatible DictConfig
        cfg = model_utils.convert_model_config_to_dict_config(cfg)
        cfg = model_utils.maybe_update_config_version(cfg)
        self.ds_class = cfg.train_ds.dataset._target_

        super().__init__(cfg=cfg, trainer=trainer)

        self.audio_to_melspec_precessor = instantiate(cfg.preprocessor)
        # We use separate preprocessor for training, because we need to pass grads and remove pitch fmax limitation
        self.trg_melspec_fn = instantiate(cfg.preprocessor, highfreq=None, use_grads=True)
        self.generator = instantiate(cfg.generator)
        self.mpd = MultiPeriodDiscriminator(debug=cfg.debug if "debug" in cfg else False)
        self.msd = MultiScaleDiscriminator(debug=cfg.debug if "debug" in cfg else False)
        self.feature_loss = FeatureMatchingLoss()
        self.discriminator_loss = DiscriminatorLoss()
        self.generator_loss = GeneratorLoss()

        self.l1_factor = cfg.get("l1_loss_factor", 45)

        self.sample_rate = self._cfg.preprocessor.sample_rate
        self.stft_bias = None

        self.input_as_mel = False
        if self._train_dl:
            self.input_as_mel = self._train_dl.dataset.load_precomputed_mel

        self.log_audio = cfg.get("log_audio", False)
        self.log_config = cfg.get("log_config", None)
        self.lr_schedule_interval = None

        # Important: this property activates manual optimization.
        self.automatic_optimization = False

    @property
    def max_steps(self):
        if "max_steps" in self._cfg:
            return self._cfg.get("max_steps")

        if "max_epochs" not in self._cfg:
            raise ValueError("Must specify 'max_steps' or 'max_epochs'.")

        if "steps_per_epoch" in self._cfg:
            return self._cfg.max_epochs * self._cfg.steps_per_epoch

        return compute_max_steps(
            max_epochs=self._cfg.max_epochs,
            accumulate_grad_batches=self.trainer.accumulate_grad_batches,
            limit_train_batches=self.trainer.limit_train_batches,
            num_workers=get_num_workers(self.trainer),
            num_samples=len(self._train_dl.dataset),
            batch_size=get_batch_size(self._train_dl),
            drop_last=self._train_dl.drop_last,
        )

    @staticmethod
    def get_warmup_steps(max_steps, warmup_steps, warmup_ratio):
        if warmup_steps is not None:
            return warmup_steps

        if warmup_ratio is not None:
            return warmup_ratio * max_steps

        return None

    def configure_optimizers(self):
        optim_config = self._cfg.optim.copy()

        OmegaConf.set_struct(optim_config, False)
        sched_config = optim_config.pop("sched", None)
        OmegaConf.set_struct(optim_config, True)

        gen_params = self.generator.parameters()
        disc_params = itertools.chain(self.msd.parameters(), self.mpd.parameters())
        optim_g = instantiate(optim_config, params=gen_params)
        optim_d = instantiate(optim_config, params=disc_params)

        if sched_config is None:
            return [optim_g, optim_d]

        max_steps = self.max_steps
        warmup_steps = self.get_warmup_steps(
            max_steps=max_steps,
            warmup_steps=sched_config.get("warmup_steps", None),
            warmup_ratio=sched_config.get("warmup_ratio", None),
        )

        OmegaConf.set_struct(sched_config, False)
        sched_config["max_steps"] = max_steps
        if warmup_steps:
            sched_config["warmup_steps"] = warmup_steps
            sched_config.pop("warmup_ratio", None)
        OmegaConf.set_struct(sched_config, True)

        scheduler_g = prepare_lr_scheduler(
            optimizer=optim_g, scheduler_config=sched_config, train_dataloader=self._train_dl
        )

        scheduler_d = prepare_lr_scheduler(
            optimizer=optim_d, scheduler_config=sched_config, train_dataloader=self._train_dl
        )

        self.lr_schedule_interval = scheduler_g["interval"]

        return [optim_g, optim_d], [scheduler_g, scheduler_d]

    def update_lr(self, interval="step"):
        schedulers = self.lr_schedulers()
        if schedulers is not None and self.lr_schedule_interval == interval:
            sch1, sch2 = schedulers
            sch1.step()
            sch2.step()

    @typecheck()
    def forward(self, *, spec):
        """
        Runs the generator, for inputs and outputs see input_types, and output_types
        """
        return self.generator(x=spec)

    @typecheck(
        input_types={"spec": NeuralType(('B', 'C', 'T'), MelSpectrogramType())},
        output_types={"audio": NeuralType(('B', 'T'), AudioSignal())},
    )
    def convert_spectrogram_to_audio(self, spec: 'torch.tensor') -> 'torch.tensor':
        return self(spec=spec).squeeze(1)

    def training_step(self, batch, batch_idx):
        audio, audio_len, audio_mel, _ = self._process_batch(batch)

        # Mel as input for L1 mel loss
        audio_trg_mel, _ = self.trg_melspec_fn(audio, audio_len)
        audio = audio.unsqueeze(1)

        audio_pred = self.generator(x=audio_mel)
        audio_pred_mel, _ = self.trg_melspec_fn(audio_pred.squeeze(1), audio_len)

        optim_g, optim_d = self.optimizers()

        # Train discriminator
        optim_d.zero_grad()
        mpd_score_real, mpd_score_gen, _, _ = self.mpd(y=audio, y_hat=audio_pred.detach())
        loss_disc_mpd, _, _ = self.discriminator_loss(
            disc_real_outputs=mpd_score_real, disc_generated_outputs=mpd_score_gen
        )
        msd_score_real, msd_score_gen, _, _ = self.msd(y=audio, y_hat=audio_pred.detach())
        loss_disc_msd, _, _ = self.discriminator_loss(
            disc_real_outputs=msd_score_real, disc_generated_outputs=msd_score_gen
        )
        loss_d = loss_disc_msd + loss_disc_mpd
        self.manual_backward(loss_d)
        optim_d.step()

        # Train generator
        optim_g.zero_grad()
        loss_mel = F.l1_loss(audio_pred_mel, audio_trg_mel)
        _, mpd_score_gen, fmap_mpd_real, fmap_mpd_gen = self.mpd(y=audio, y_hat=audio_pred)
        _, msd_score_gen, fmap_msd_real, fmap_msd_gen = self.msd(y=audio, y_hat=audio_pred)
        loss_fm_mpd = self.feature_loss(fmap_r=fmap_mpd_real, fmap_g=fmap_mpd_gen)
        loss_fm_msd = self.feature_loss(fmap_r=fmap_msd_real, fmap_g=fmap_msd_gen)
        loss_gen_mpd, _ = self.generator_loss(disc_outputs=mpd_score_gen)
        loss_gen_msd, _ = self.generator_loss(disc_outputs=msd_score_gen)
        loss_g = loss_gen_msd + loss_gen_mpd + loss_fm_msd + loss_fm_mpd + loss_mel * self.l1_factor
        self.manual_backward(loss_g)
        optim_g.step()

        self.update_lr()

        metrics = {
            "g_loss_fm_mpd": loss_fm_mpd,
            "g_loss_fm_msd": loss_fm_msd,
            "g_loss_gen_mpd": loss_gen_mpd,
            "g_loss_gen_msd": loss_gen_msd,
            "g_loss": loss_g,
            "d_loss_mpd": loss_disc_mpd,
            "d_loss_msd": loss_disc_msd,
            "d_loss": loss_d,
            "global_step": self.global_step,
            "lr": optim_g.param_groups[0]['lr'],
        }
        self.log_dict(metrics, on_step=True, sync_dist=True)
        self.log("g_l1_loss", loss_mel, prog_bar=True, logger=False, sync_dist=True)

    def on_train_epoch_end(self) -> None:
        self.update_lr("epoch")

    def validation_step(self, batch, batch_idx):
        audio, audio_len, audio_mel, audio_mel_len = self._process_batch(batch)

        audio_pred = self(spec=audio_mel)

        if self.input_as_mel:
            gt_mel, gt_mel_len = self.audio_to_melspec_precessor(audio, audio_len)
        audio_pred_mel, _ = self.audio_to_melspec_precessor(audio_pred.squeeze(1), audio_len)
        loss_mel = F.l1_loss(audio_mel, audio_pred_mel)

        self.log_dict({"val_loss": loss_mel}, on_epoch=True, sync_dist=True)

        # Plot audio once per epoch
        if self.log_audio and batch_idx == 0 and isinstance(self.logger, WandbLogger) and HAVE_WANDB:
            # Perform bias denoising
            pred_denoised = self._bias_denoise(audio_pred, audio_mel).squeeze(1)
            pred_denoised_mel, _ = self.audio_to_melspec_precessor(pred_denoised, audio_len)

            clips = []
            specs = []
            for i in range(min(5, audio.shape[0])):
                clips += [
                    wandb.Audio(
                        audio[i, : audio_len[i]].data.cpu().numpy(),
                        caption=f"real audio {i}",
                        sample_rate=self.sample_rate,
                    ),
                    wandb.Audio(
                        audio_pred[i, 0, : audio_len[i]].data.cpu().numpy().astype('float32'),
                        caption=f"generated audio {i}",
                        sample_rate=self.sample_rate,
                    ),
                    wandb.Audio(
                        pred_denoised[i, : audio_len[i]].data.cpu().numpy(),
                        caption=f"denoised audio {i}",
                        sample_rate=self.sample_rate,
                    ),
                ]
                specs += [
                    wandb.Image(
                        plot_spectrogram_to_numpy(audio_mel[i, :, : audio_mel_len[i]].data.cpu().numpy()),
                        caption=f"input mel {i}",
                    ),
                    wandb.Image(
                        plot_spectrogram_to_numpy(audio_pred_mel[i, :, : audio_mel_len[i]].data.cpu().numpy()),
                        caption=f"output mel {i}",
                    ),
                    wandb.Image(
                        plot_spectrogram_to_numpy(pred_denoised_mel[i, :, : audio_mel_len[i]].data.cpu().numpy()),
                        caption=f"denoised mel {i}",
                    ),
                ]
                if self.input_as_mel:
                    specs += [
                        wandb.Image(
                            plot_spectrogram_to_numpy(gt_mel[i, :, : audio_mel_len[i]].data.cpu().numpy()),
                            caption=f"gt mel {i}",
                        ),
                    ]

            self.logger.experiment.log({"audio": clips, "specs": specs})

    def _process_batch(self, batch):
        if self.input_as_mel:
            audio, audio_len, audio_mel = batch
            audio_mel_len = [audio_mel.shape[1]] * audio_mel.shape[0]
            return audio, audio_len, audio_mel, audio_mel_len

        if self.ds_class == "nemo.collections.tts.data.vocoder_dataset.VocoderDataset":
            audio = batch.get("audio")
            audio_len = batch.get("audio_lens")
        else:
            audio, audio_len = batch

        audio_mel, audio_mel_len = self.audio_to_melspec_precessor(audio, audio_len)
        return audio, audio_len, audio_mel, audio_mel_len

    def _bias_denoise(self, audio, mel):
        def stft(x):
            comp = torch.stft(x.squeeze(1), n_fft=1024, hop_length=256, win_length=1024, return_complex=True)
            comp = torch.view_as_real(comp)
            real, imag = comp[..., 0], comp[..., 1]
            mags = torch.sqrt(real**2 + imag**2)
            phase = torch.atan2(imag, real)
            return mags, phase

        def istft(mags, phase):
            comp = torch.stack([mags * torch.cos(phase), mags * torch.sin(phase)], dim=-1)
            x = torch.istft(torch.view_as_complex(comp), n_fft=1024, hop_length=256, win_length=1024)
            return x

        # Create bias tensor
        if self.stft_bias is None or self.stft_bias.shape[0] != audio.shape[0]:
            audio_bias = self(spec=torch.zeros_like(mel, device=mel.device))
            self.stft_bias, _ = stft(audio_bias)
            self.stft_bias = self.stft_bias[:, :, 0][:, :, None]

        audio_mags, audio_phase = stft(audio)
        audio_mags = audio_mags - self.cfg.get("denoise_strength", 0.0025) * self.stft_bias
        audio_mags = torch.clamp(audio_mags, 0.0)
        audio_denoised = istft(audio_mags, audio_phase).unsqueeze(1)

        return audio_denoised

    def _setup_train_dataloader(self, cfg):
        dataset = instantiate(cfg.dataset)
        sampler = dataset.get_sampler(cfg.dataloader_params.batch_size, world_size=self.trainer.world_size)
        data_loader = torch.utils.data.DataLoader(
            dataset, collate_fn=dataset.collate_fn, sampler=sampler, **cfg.dataloader_params
        )
        return data_loader

    def _setup_test_dataloader(self, cfg):
        dataset = instantiate(cfg.dataset)
        data_loader = torch.utils.data.DataLoader(dataset, collate_fn=dataset.collate_fn, **cfg.dataloader_params)
        return data_loader

    def __setup_dataloader_from_config(self, cfg, shuffle_should_be: bool = True, name: str = "train"):
        if "dataset" not in cfg or not isinstance(cfg.dataset, DictConfig):
            raise ValueError(f"No dataset for {name}")
        if "dataloader_params" not in cfg or not isinstance(cfg.dataloader_params, DictConfig):
            raise ValueError(f"No dataloader_params for {name}")
        if shuffle_should_be:
            if 'shuffle' not in cfg.dataloader_params:
                logging.warning(
                    f"Shuffle should be set to True for {self}'s {name} dataloader but was not found in its "
                    "config. Manually setting to True"
                )
                with open_dict(cfg["dataloader_params"]):
                    cfg.dataloader_params.shuffle = True
            elif not cfg.dataloader_params.shuffle:
                logging.error(f"The {name} dataloader for {self} has shuffle set to False!!!")
        elif not shuffle_should_be and cfg.dataloader_params.shuffle:
            logging.error(f"The {name} dataloader for {self} has shuffle set to True!!!")

        dataset = instantiate(cfg.dataset)
        return torch.utils.data.DataLoader(dataset, collate_fn=dataset.collate_fn, **cfg.dataloader_params)

    def setup_training_data(self, cfg):
        if self.ds_class == "nemo.collections.tts.data.vocoder_dataset.VocoderDataset":
            self._train_dl = self._setup_train_dataloader(cfg)
        else:
            self._train_dl = self.__setup_dataloader_from_config(cfg)

    def setup_validation_data(self, cfg):
        if self.ds_class == "nemo.collections.tts.data.vocoder_dataset.VocoderDataset":
            self._validation_dl = self._setup_test_dataloader(cfg)
        else:
            self._validation_dl = self.__setup_dataloader_from_config(cfg, shuffle_should_be=False, name="validation")

    def setup_test_data(self, cfg):
        pass

    def configure_callbacks(self):
        if not self.log_config:
            return []

        sample_ds_class = self.log_config.dataset._target_
        if sample_ds_class != "nemo.collections.tts.data.vocoder_dataset.VocoderDataset":
            raise ValueError(f"Sample logging only supported for VocoderDataset, got {sample_ds_class}")

        data_loader = self._setup_test_dataloader(self.log_config)
        generators = instantiate(self.log_config.generators)
        log_dir = Path(self.log_config.log_dir) if self.log_config.log_dir else None
        log_callback = LoggingCallback(
            generators=generators,
            data_loader=data_loader,
            log_epochs=self.log_config.log_epochs,
            epoch_frequency=self.log_config.epoch_frequency,
            output_dir=log_dir,
            loggers=self.trainer.loggers,
            log_tensorboard=self.log_config.log_tensorboard,
            log_wandb=self.log_config.log_wandb,
        )

        return [log_callback]

    @classmethod
    def list_available_models(cls) -> 'Optional[Dict[str, str]]':
        list_of_models = []
        model = PretrainedModelInfo(
            pretrained_model_name="tts_en_hifigan",
            location="https://api.ngc.nvidia.com/v2/models/nvidia/nemo/tts_hifigan/versions/1.0.0rc1/files/tts_hifigan.nemo",
            description="This model is trained on LJSpeech audio sampled at 22050Hz and mel spectrograms generated from"
            " Tacotron2, TalkNet, and FastPitch. This model has been tested on generating female English "
            "voices with an American accent.",
            class_=cls,
        )
        list_of_models.append(model)

        model = PretrainedModelInfo(
            pretrained_model_name="tts_en_lj_hifigan_ft_mixertts",
            location="https://api.ngc.nvidia.com/v2/models/nvidia/nemo/tts_en_lj_hifigan/versions/1.6.0/files/tts_en_lj_hifigan_ft_mixertts.nemo",
            description="This model is trained on LJSpeech audio sampled at 22050Hz and mel spectrograms generated from"
            " Mixer-TTS. This model has been tested on generating female English voices with an American accent.",
            class_=cls,
        )
        list_of_models.append(model)

        model = PretrainedModelInfo(
            pretrained_model_name="tts_en_lj_hifigan_ft_mixerttsx",
            location="https://api.ngc.nvidia.com/v2/models/nvidia/nemo/tts_en_lj_hifigan/versions/1.6.0/files/tts_en_lj_hifigan_ft_mixerttsx.nemo",
            description="This model is trained on LJSpeech audio sampled at 22050Hz and mel spectrograms generated from"
            " Mixer-TTS-X. This model has been tested on generating female English voices with an American accent.",
            class_=cls,
        )
        list_of_models.append(model)

        model = PretrainedModelInfo(
            pretrained_model_name="tts_en_hifitts_hifigan_ft_fastpitch",
            location="https://api.ngc.nvidia.com/v2/models/nvidia/nemo/tts_en_multispeaker_fastpitchhifigan/versions/1.10.0/files/tts_en_hifitts_hifigan_ft_fastpitch.nemo",
            description="This model is trained on HiFiTTS audio sampled at 44100Hz and mel spectrograms generated from"
            " FastPitch. This model has been tested on generating male and female English voices with an American accent.",
            class_=cls,
        )
        list_of_models.append(model)

        # de-DE, single male speaker, 22050 Hz, Thorsten Müller’s German Neutral-TTS Dataset, 21.02
        model = PretrainedModelInfo(
            pretrained_model_name="tts_de_hifigan_singleSpeaker_thorstenNeutral_2102",
            location="https://api.ngc.nvidia.com/v2/models/nvidia/nemo/tts_de_fastpitchhifigan/versions/1.15.0/files/tts_de_hifigan_thorstens2102.nemo",
            description="This model is finetuned from the HiFiGAN pretrained checkpoint `tts_en_lj_hifigan_ft_mixerttsx`"
            " by the mel-spectrograms generated from the FastPitch checkpoint `tts_de_fastpitch_singleSpeaker_thorstenNeutral_2102`."
            " This model has been tested on generating male German neutral voices.",
            class_=cls,
        )
        list_of_models.append(model)

        # de-DE, single male speaker, 22050 Hz, Thorsten Müller’s German Neutral-TTS Dataset, 22.10
        model = PretrainedModelInfo(
            pretrained_model_name="tts_de_hifigan_singleSpeaker_thorstenNeutral_2210",
            location="https://api.ngc.nvidia.com/v2/models/nvidia/nemo/tts_de_fastpitchhifigan/versions/1.15.0/files/tts_de_hifigan_thorstens2210.nemo",
            description="This model is finetuned from the HiFiGAN pretrained checkpoint `tts_en_lj_hifigan_ft_mixerttsx`"
            " by the mel-spectrograms generated from the FastPitch checkpoint `tts_de_fastpitch_singleSpeaker_thorstenNeutral_2210`."
            " This model has been tested on generating male German neutral voices.",
            class_=cls,
        )
        list_of_models.append(model)

        # de-DE, multi-speaker, 5 speakers, 44100 Hz, HUI-Audio-Corpus-German Clean.
        model = PretrainedModelInfo(
            pretrained_model_name="tts_de_hui_hifigan_ft_fastpitch_multispeaker_5",
            location="https://api.ngc.nvidia.com/v2/models/nvidia/nemo/tts_de_fastpitch_multispeaker_5/versions/1.11.0/files/tts_de_hui_hifigan_ft_fastpitch_multispeaker_5.nemo",
            description="This model is finetuned from the HiFiGAN pretrained checkpoint `tts_en_hifitts_hifigan_ft_fastpitch` "
            "by the mel-spectrograms generated from the FastPitch checkpoint `tts_de_fastpitch_multispeaker_5`. This model "
            "has been tested on generating male and female German voices.",
            class_=cls,
        )
        list_of_models.append(model)

        # Spanish, multi-speaker, 44100 Hz, Latin American Spanish OpenSLR
        model = PretrainedModelInfo(
            pretrained_model_name="tts_es_hifigan_ft_fastpitch_multispeaker",
            location="https://api.ngc.nvidia.com/v2/models/nvidia/nemo/tts_es_multispeaker_fastpitchhifigan/versions/1.15.0/files/tts_es_hifigan_ft_fastpitch_multispeaker.nemo",
            description="This model is trained on the audio from 6 crowdsourced Latin American Spanish OpenSLR "
            "datasets and finetuned on the mel-spectrograms generated from the FastPitch checkpoint "
            "`tts_es_fastpitch_multispeaker`. This model has been tested on generating male and female "
            "Spanish voices with Latin American accents.",
            class_=cls,
        )
        list_of_models.append(model)

        # zh, single female speaker, 22050Hz, SFSpeech Bilingual Chinese/English dataset, improved model using richer
        # dict and jieba word segmenter for polyphone disambiguation.
        model = PretrainedModelInfo(
            pretrained_model_name="tts_zh_hifigan_sfspeech",
            location="https://api.ngc.nvidia.com/v2/models/nvidia/nemo/tts_zh_fastpitch_hifigan_sfspeech/versions/1.15.0/files/tts_zh_hifigan_sfspeech.nemo",
            description="This model is finetuned from the HiFiGAN pretrained checkpoint `tts_en_lj_hifigan_ft_mixerttsx`"
            " by the mel-spectrograms generated from the FastPitch checkpoint `tts_zh_fastpitch_sfspeech`."
            " This model has been tested on generating female Mandarin Chinese voices.",
            class_=cls,
        )
        list_of_models.append(model)

        return list_of_models

    def load_state_dict(self, state_dict, strict=True):
        # Override load_state_dict to give us some flexibility to be backward-compatible with old checkpoints
        new_state_dict = {}
        num_resblocks = len(self.cfg['generator']['resblock_kernel_sizes'])
        for k, v in state_dict.items():
            new_k = k
            if 'resblocks' in k:
                parts = k.split(".")
                # only do this is the checkpoint type is older
                if len(parts) == 6:
                    layer = int(parts[2])
                    new_layer = f"{layer // num_resblocks}.{layer % num_resblocks}"
                    new_k = f"generator.resblocks.{new_layer}.{'.'.join(parts[3:])}"
            new_state_dict[new_k] = v
        super().load_state_dict(new_state_dict, strict=strict)

    # Methods for model exportability
    def _prepare_for_export(self, **kwargs):
        if self.generator is not None:
            try:
                self.generator.remove_weight_norm()
            except ValueError:
                return

    @property
    def input_types(self):
        return {
            "spec": NeuralType(('B', 'D', 'T'), MelSpectrogramType()),
        }

    @property
    def output_types(self):
        return {
            "audio": NeuralType(('B', 'S', 'T'), AudioSignal(self.sample_rate)),
        }

    def input_example(self, max_batch=1, max_dim=256):
        """
        Generates input examples for tracing etc.
        Returns:
            A tuple of input examples.
        """
        par = next(self.parameters())
        mel = torch.randn((max_batch, self.cfg['preprocessor']['nfilt'], max_dim), device=self.device, dtype=par.dtype)
        return ({'spec': mel},)

    def forward_for_export(self, spec):
        """
        Runs the generator, for inputs and outputs see input_types, and output_types
        """
        return self.generator(x=spec)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Callable, Optional


@dataclass
class MultiModalToken:
    """
    Base class for multimodal tokens representing different media types.
    """

    token_str: str
    token_index: int
    media_type: str
    use_start_end: bool
    encoder_fn: Optional[Callable] = None


@dataclass
class Qwen2VLImageToken(MultiModalToken):
    """Image Token class"""

    token_str: str = "<|image_pad|>"
    token_index: int = -200
    media_type: str = "image"
    use_start_end: bool = False


@dataclass
class Qwen2VLVideoToken(MultiModalToken):
    """Video Token class"""

    token_str: str = "<|video_pad|>"
    token_index: int = -300
    media_type: str = "video"
    use_start_end: bool = False


# Constants for token indexing and special token mapping
IGNORE_INDEX = -100
IMAGE_TOKEN_INDEX = Qwen2VLImageToken.token_index
VIDEO_TOKEN_INDEX = Qwen2VLVideoToken.token_index
OBJECT_REF_START_TOKEN_INDEX = 151646
OBJECT_REF_END_TOKEN_INDEX = 151647
BOX_START_TOKEN_INDEX = 151648
BOX_END_TOKEN_INDEX = 151649
QUAD_START_TOKEN_INDEX = 151650
QUAD_END_TOKEN_INDEX = 151651
VISION_START_TOKEN_INDEX = 151652
VISION_END_TOKEN_INDEX = 151653
PAD_TOKEN_INDEX = 151643
HF_IMAGE_TOKEN_INDEX = 151655
HF_VIDEO_TOKEN_INDEX = 151656

SPECIAL_TOKEN_MAP = [
    (Qwen2VLImageToken.token_str, Qwen2VLImageToken.token_index),
    (Qwen2VLVideoToken.token_str, Qwen2VLVideoToken.token_index),
    ("<|object_ref_start|>", OBJECT_REF_START_TOKEN_INDEX),
    ("<|object_ref_end|>", OBJECT_REF_END_TOKEN_INDEX),
    ("<|box_start|>", BOX_START_TOKEN_INDEX),
    ("<|box_end|>", BOX_END_TOKEN_INDEX),
    ("<|quad_start|>", QUAD_START_TOKEN_INDEX),
    ("<|quad_end|>", QUAD_END_TOKEN_INDEX),
    ("<|vision_start|>", VISION_START_TOKEN_INDEX),
    ("<|vision_end|>", VISION_END_TOKEN_INDEX),
]



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .duplex_ear_tts import DuplexEARTTS
from .duplex_s2s_model import DuplexS2SModel
from .duplex_s2s_speech_decoder_model import DuplexS2SSpeechDecoderModel
from .salm import SALM
from .salm_asr_decoder import SALMWithAsrDecoder

__all__ = [
    'DuplexS2SModel',
    'DuplexS2SSpeechDecoderModel',
    'DuplexEARTTS',
    'SALM',
    'SALMWithAsrDecoder',
]
